#!/usr/bin/env python

import sys
import logging


import math
import numpy as np
import pandas as pd
from numpy import linalg as LA

from collections import defaultdict
from collections import Counter
from CodeEntropy.FunctionCollection import Utils

nested_dict = lambda: defaultdict(nested_dict)
        #create nested dict in one go



def processEE(num_frames, totFrames, Aclass, solvent, waterTuple, 
        temperature, level, name, forceUnits, verbosePrint):
    '''
    Populate and process Smix and Sor into global dictionaries
    '''

    verbosePrint('\n\n---ANALYSIS_TYPE_%s---\n' % (name))
    verbosePrint('\n\n---ORIENTATIONAL_ENTROPY_CALCULATIONS---\n')
    if any(i in solvent for i in waterTuple) == True:
        SorCalculation(Aclass, 'Sor_test2', level, waterTuple, verbosePrint)


    verbosePrint('\n\n --UA_VIBRATIONAL_ENTROPY--\n')
    SvibCalculations(totFrames, Aclass, temperature, forceUnits, verbosePrint)
    verbosePrint('\n\n --WM_VIBRATIONAL_ENTROPY--\n')
    WMSvibCalculations(Aclass, totFrames, temperature, 
            forceUnits, verbosePrint)

    verbosePrint('\n\n --WM_CONFORMATIONAL_ENTROPY(DIH)--\n')
    process_dihedrals(Aclass, verbosePrint)


    verbosePrint('\n\n--PRINT_ALL_VARIABLES--\n')
    # solventData, soluteData = saveAllVariables(num_frames, Aclass, level, name, solvent, 
    #         waterTuple, verbosePrint)
    solventData, soluteData = saveAllVariables(num_frames, Aclass, level, name, solvent, 
             waterTuple, verbosePrint)

    if level in ['residLevel_resname', 'atomLevel']:
        verbosePrint('\n\n--CONTACTS--\n')
        # contactMatrix = contactCalculation(Aclass, level, totFrames, verbosePrint)
        contactMatrix = contactCalculation(Aclass, level, totFrames, verbosePrint)

        return {'solventData':solventData, 'soluteData':soluteData, 'contactMatrix':contactMatrix}
    
    return {'solventData':solventData, 'soluteData':soluteData}



def contactCalculation(Aclass, level, totFrames, verbosePrint):
    '''
    resid matrix of RAD contacts between resids that are not water and 
    the nearest non-like to it.
    '''

    verbosePrint('\n*******_RESID_CONTACTS_**********\n')
    # data = open('resid_contact_matrix_%s.csv' % (level), 'w')
    contact_matrix_name = f'resid_contact_matrix_{level}.csv'
    if level == 'atomLevel':
        # data.write('\n'.join(['centre_resid,neighbour_resid,count,'\
        #         'centre_resname,neighbour_resname,centre_atom,'\
        #         'neighbour_atom']) 
        #         + '\n')
        Utils.printOut(contact_matrix_name, 'centre_resid,neighbour_resid,count,centre_resname,neighbour_resname,centre_atom,neighbour_atom')
        contactMatrix = pd.DataFrame(columns=['centre_resid','neighbour_resid','count','centre_resname','neighbour_resname','centre_atom', 'neighbour_atom'])
    else:
        # data.write('\n'.join(['centre_resid,neighbour_resid,count,'\
        #         'centre_resname,neighbour_resname']) 
        #         + '\n') 
        Utils.printOut(contact_matrix_name, 'centre_resid,neighbour_resid,count,centre_resname,neighbour_resname')
        contactMatrix = pd.DataFrame(columns=['centre_resid','neighbour_resid','count','centre_resname','neighbour_resname'])
    for centre_info, neighbour_resid_key in \
            sorted(list(Aclass.resid_contact_matrix_dict.items())):
        for neighbour_info, count in \
                sorted(list(neighbour_resid_key.items())):
            #verbosePrint(centre_info, neighbour_info, count)
            if count > 0:
                if level == 'atomLevel':
                    centre_atom = centre_info[0]
                    centre_resname = centre_info[1]
                    centre_resid = centre_info[2]
                    neighbour_atom = neighbour_info[0]
                    neighbour_resname = neighbour_info[1]
                    neighbour_resid = neighbour_info[2]
                    count = float(count) / float(totFrames)
                    # data.write('\n'.join(['%s,%s,%s,%s,%s,%s,%s' 
                    #         % (centre_resid, 
                    #         neighbour_resid, count, centre_resname, 
                    #         neighbour_resname, centre_atom, 
                    #         neighbour_atom)]) + '\n')
                    Utils.printOut(contact_matrix_name, f"{centre_resid},{neighbour_resid},{count},{centre_resname},{neighbour_resname},{centre_atom},{neighbour_atom}")
                    newRowContact = pd.DataFrame({'centre_resid': centre_resid,
                                            'neighbour_resid':neighbour_resid,
                                            'count': count,
                                            'centre_resname': centre_resname,
                                            'neighbour_resname': neighbour_resname,
                                            'centre_atom': centre_atom,
                                            'neighbour_atom': neighbour_atom}, index=[0])
                    contactMatrix = pd.concat([contactMatrix, newRowContact], ignore_index=True)
                else:
                    centre_resname = centre_info[0]
                    centre_resid = centre_info[1]
                    neighbour_resname = neighbour_info[0]
                    neighbour_resid = neighbour_info[1]
                    count = float(count) / float(totFrames)
                    # data.write('\n'.join(['%s,%s,%s,%s,%s' % (centre_resid, 
                    #         neighbour_resid, count, centre_resname, 
                    #         neighbour_resname)]) + '\n')
                    Utils.printOut(contact_matrix_name, f"{centre_resid},{neighbour_resid},{count},{centre_resname},{neighbour_resname}")
                    newRowContact = pd.DataFrame({'centre_resid': centre_resid,
                                            'neighbour_resid':neighbour_resid,
                                            'count': count,
                                            'centre_resname': centre_resname,
                                            'neighbour_resname': neighbour_resname}, index=[0])
                    contactMatrix = pd.concat([contactMatrix, newRowContact], ignore_index=True)
            else:
                continue


    # data.close()
    return contactMatrix



def SorCalculation(Aclass, methodType, level, waterTuple, verbosePrint):
    '''
    '''

    por_dict = nested_dict()
    shells_ppD_ppA_dict = nested_dict()
    verbosePrint('\n\n__ORIENTATIONAL_ENTROPY_%s' % (methodType))
    num_molecules0 = sum(count
            for atomType in Aclass.RADshell_Sor_dict.values()
            #for Nc in atomType.values() 
            for RADshell_num in atomType.values() 
            for RAD_str in RADshell_num.values()
            for ADs in RAD_str.values()
            for key, HBcount in ADs.items() 
            for HB, count in HBcount.items()
            if key == 'D')
    verbosePrint('\nall molecule count:', num_molecules0)
    S1 = 0
    for nearest, assigned_key in \
            sorted(list(Aclass.RADshell_Sor_dict.items())):
        num_molecules1 = sum(count
                #for Nc in assigned_key.values() 
                for RADshell_num in assigned_key.values() 
                for RAD_str in RADshell_num.values()
                for ADs in RAD_str.values()
                for AD, HBcount in ADs.items() 
                for HB, count in HBcount.items()
                if AD == 'D')

        nearest_resname = nearest.split('_')[0]
        verbosePrint('\n\nnearest:', nearest, 'count:', num_molecules1, 
                '/', num_molecules0)

        S2 = 0
        for assigned, shell_num_key in sorted(list(assigned_key.items())):
            num_molecules2 = sum(count
                    #for RADshell_num in Nc_key.values() 
                    for RAD_str in shell_num_key.values()
                    for ADs in RAD_str.values()
                    for AD, HBcount in ADs.items() 
                    for HB, count in HBcount.items()
                    if AD == 'D')

            assigned_resname = assigned[0].split('_')[0]
            verbosePrint('\tassigned:', assigned, 'count:', num_molecules2,
                    '/', num_molecules1)
            N_H = assigned[1]
            verbosePrint('\tN_H', N_H)
            S4 = 0
            for shell_num, RADlist_key in \
                    sorted(list(shell_num_key.items())):

                num_molecules4 = sum(count
                        for ADs in RADlist_key.values()
                        for AD, HBcount in ADs.items() 
                        for HB, count in HBcount.items()
                        if AD == 'D')

                verbosePrint('\t\t\tshell_num:', shell_num[0], 'count:', 
                        num_molecules4, '/', num_molecules2)
                S5 = 0
                S_shell = 0
                ave_pbias_ave = 0
                ave_Nc_eff = 0

                for RADlist, AD_key in \
                        sorted(list(RADlist_key.items())):
                    verbosePrint('\t'*4, '-----RADshell------')
                    num_molecules5 = sum(count
                            for AD, HBcount in AD_key.items() 
                            for HB, count in HBcount.items()
                            if AD == 'D')

                    ## dict for p(c) log p(c)
                    if methodType == 'Sor_test2' and \
                            level == 'moleculeLevel' and \
                            assigned[0] in waterTuple and \
                            int(shell_num[0]) == 1:
                        newRADlist = []
                        for n in RADlist:
                            if n == '0':
                                newRADlist.append(nearest_resname)
                            else:
                                newRADlist.append(n)
                        newRADlist = tuple(sorted(newRADlist))
                        try:
                            if newRADlist not in \
                                    Aclass.p_coord_dict[len(RADlist)]:
                                Aclass.p_coord_dict[len(RADlist)]\
                                        [newRADlist] = 0
                            Aclass.p_coord_dict[len(RADlist)][newRADlist] += \
                                    num_molecules5
                        except AttributeError:
                            pass

                    RAD_l = ('%s' % (','.join(map(str, RADlist))))
                    verbosePrint('\t\t\t\tneighbours:', RAD_l, 'Nc:', 
                            len(RADlist), 'count:', 
                            num_molecules5, '/', num_molecules4)
                    S6 = 0
                    Sor_list = []
                    pbias_ave = 0
                    Nc_eff = 0


                    if methodType == 'Sor_test2':
                        D_values = None
                        A_values = None
                        for AD, values in \
                                sorted(list(AD_key.items())):
                            if AD == 'D':
                                D_values = values
                            if AD == 'A':
                                A_values = values
                        verbosePrint('\t'*4, '* SEPARATE ADs *')
                        S7, pbias_ave, Nc_eff, ppD_ppA_dict = \
                                RAD_or_entropyCalc_test2(
                                RADlist, D_values,
                                A_values, N_H, shell_num[1], 
                                num_molecules5, nearest_resname, 
                                waterTuple, verbosePrint)

                        ### ppD_ppA_dict population
                        for c in range(0, int(num_molecules5)):
                            for i, pp in ppD_ppA_dict.items():
                                ppD = pp[0]
                                ppA = pp[1]
                                if i not in shells_ppD_ppA_dict[nearest]\
                                        [assigned]\
                                        [float(shell_num[0])]:
                                    shells_ppD_ppA_dict[nearest][assigned]\
                                            [float(shell_num[0])]\
                                            [i] = [0, 0, 0]
                                ppD_stored = shells_ppD_ppA_dict[nearest]\
                                        [assigned][float(shell_num[0])][i][0]
                                ppA_stored = shells_ppD_ppA_dict[nearest]\
                                        [assigned][float(shell_num[0])][i][1]
                                count_stored = shells_ppD_ppA_dict[nearest]\
                                        [assigned][float(shell_num[0])][i][2]


                                count_add = count_stored + 1
                                ppD_add = ((ppD + ppD_stored * 
                                        count_stored) / float(count_add))
                                ppA_add = ((ppA + ppA_stored * 
                                        count_stored) / float(count_add))

                                shells_ppD_ppA_dict[nearest][assigned]\
                                        [float(shell_num[0])]\
                                        [i][0] = ppD_add
                                shells_ppD_ppA_dict[nearest][assigned]\
                                        [float(shell_num[0])]\
                                        [i][1] = ppA_add
                                shells_ppD_ppA_dict[nearest][assigned]\
                                        [float(shell_num[0])]\
                                        [i][2] = count_add


                                            
                        Sor_list.append([S7, 'DA'])
                        verbosePrint('\t\t\t\tS7_DA = ', round(S7, 3))


                    Sor_list = sorted(Sor_list)
                    verbosePrint()
                    verbosePrint('\t'*4, '*** Sor-list:')
                    c = 0
                    for s, label in Sor_list:
                        c += 1
                        if c == 1:
                            verbosePrint('\t'*4, round(s, 3), label, 
                                    '<-- selected Sor')
                        else:
                            verbosePrint('\t'*4, round(s, 3), label)
                    #### use smallest Sor as selected value
                    S6 += (Sor_list[0][0] * float(num_molecules5) / 
                        float(num_molecules4))
                    S_shell += (Sor_list[0][0] * 
                            float(num_molecules5) / 
                            float(num_molecules4))
                    ave_pbias_ave += (pbias_ave * 
                            float(num_molecules5) / 
                            float(num_molecules4))

                    ave_Nc_eff += (Nc_eff * 
                            float(num_molecules5) / 
                            float(num_molecules4))

                    if methodType == 'Sor_test2':
                        Aclass.Sor_reference_dict[nearest]\
                                [assigned[0]]\
                                [str(shell_num[0])][RADlist] = \
                                Sor_list[0][0] * 8.314
                        #print(RADlist, Sor_list[0][0] * 8.314)

                    verbosePrint('\t\t\t\tS6 = ', round(S6, 3))
                    verbosePrint()
                    S5 += (S6 * float(num_molecules4) /
                    float(num_molecules2))




                Aclass.allVariables_dict[nearest][assigned]\
                        [float(shell_num[0])]['%s' % (methodType)] = \
                        S_shell * 8.314, num_molecules4

                '''
                if methodType == 'Sor_test2':
                    Aclass.allVariables_dict[nearest][assigned]\
                            [float(shell_num[0])]\
                            ['pbias_ave_%s' % (methodType)] = \
                            ave_pbias_ave, num_molecules4

                    Aclass.allVariables_dict[nearest][assigned]\
                            [float(shell_num[0])]\
                            ['Nc_eff_%s' % (methodType)] = \
                            ave_Nc_eff, num_molecules4
                '''

                verbosePrint('\t\t\tS5 = ', round(S5, 3))
                S4 += (S5 * float(num_molecules4) /
                float(num_molecules2))

            verbosePrint('\t\tS4 = ', round(S4, 3))
            S2 += (S4 * float(num_molecules1) /
            float(num_molecules0))

        verbosePrint('\tS2 = ', round(S2, 3))
        S1 += S2
    verbosePrint('S1 = ', round(S1, 3))


    if methodType == 'Sor_test2':
        verbosePrint('\n\n__ppX_ORIENTATIONAL_ENTROPY_%s' % (methodType))
        verbosePrint('nearest,assigned,shell_num,i,ppD,ppA,count')
        for nearest, assigned_key in sorted(list(shells_ppD_ppA_dict.items())):
            for assigned, shell_num_key in sorted(list(assigned_key.items())):
                for shell_num, i_key in sorted(list(shell_num_key.items())):
                    for i, pp in sorted(list(i_key.items())):
                        ppD, ppA, count = pp[0], pp[1], pp[2]
                        verbosePrint('%s,%s,%s,%s,%s,%s,%s' % 
                                (nearest, assigned[0], int(shell_num), 
                                i, round(ppD, 5), round(ppA, 5), count))




def RAD_or_entropyCalc_test2(shell, shellD, shellA, N_H, shell_num, 
        referenceCount, nearest, waterTuple, verbosePrint):
    '''
    ***** S_or
    '''

    ####### get info about shell
    shellCount = Counter([str(N) for N in shell]) 
        #count constituents in shell
    Nc = sum(shellCount.values())
    Nc_solvent_only = 0
    for n in shell:
        if n == 'X':
            Nc_solvent_only += 1
        else:
            try:
                n = int(n)
                if n != 0:
                    Nc_solvent_only += 1
                else:
                    continue
            except ValueError:
                pass


    soluteA_pi, soluteD_pi = None, None
    pA_degen_dict, pD_degen_dict = None, None
    for AD, shellDs in [['A', shellA], ['D', shellD]]:
        verbosePrint('\t'*5, '__%s__' % (AD))
        ####### get info about shell
        shellCount = Counter([str(N) for N in shell]) 
            #count constituents in shell
        Dcount = Counter([tuple(Ds) for Ds in shellDs])
        num_orients = len(shellDs)

        solute_donor = False
        solvent_donor = False

        for ds in shellDs:
            for d in ds:
                try:
                    d = int(d)
                    if d == 0:
                        solute_donor = True
                    else:
                        continue
                except ValueError:
                    if d in ['X', 'H']:
                        solvent_donor = True
                    else:
                        solute_donor = True


        ####### pi only, degeneracy info for each orientation type observed
        i_counts_dict = {}
        tot_i_count = 0
        for donors, count in Dcount.items():
            for i in donors:
                if i not in i_counts_dict:
                    i_counts_dict[i]  = 0
                i_counts_dict[i] += count
                tot_i_count += count

        verbosePrint('\t'*5, '--i--')
        for i, count in i_counts_dict.items():
            verbosePrint('\t'*5, i, count)
        verbosePrint('\t'*5, 'sum i counts:', tot_i_count)

        pi_degen_dict = {}
        solute_pis = 0
        largest_pi = 0
        ww_Ds = 0
        for i, count in i_counts_dict.items():
            degeneracy = shellCount[i]
            if degeneracy == 0:
                degeneracy = 1
            pi = (float(count) / float(tot_i_count)) / float(degeneracy)
            pi_degen_dict[i] = [pi, degeneracy, count]
            verbosePrint('\t'*6, 'pi', i, round(pi, 4), 
                    'degen:', degeneracy, 
                    'pi_d:', round((pi * degeneracy), 4))
            #verbosePrint(i, pi)
            if pi * degeneracy > largest_pi:
                largest_pi = pi * degeneracy

            solute_pi, solvent_pi = False, False
            try:
                i = int(i)
                if i == 0 and nearest not in waterTuple:
                    solute_pi = True
                else:
                    ww_Ds += count
                    continue
            except ValueError:
                if i in ['X', 'H']:
                    solvent_pi = True
                    ww_Ds += count
                else:
                    solute_pi = True
            if solute_pi == True:
                solute_pis += pi * degeneracy



        ####### calculate number of observed orientations
        #pi_max = max(map(lambda x: x[0], pi_degen_dict.values()))
        #verbosePrint('\t'*6, 'pi_max', round(pi_max, 5))

        #verbosePrint(ww_Ds, referenceCount)
        per_w_Ds = ww_Ds / float(referenceCount)


        if AD == 'A':
            pA_degen_dict = pi_degen_dict
        if AD == 'D':
            pD_degen_dict = pi_degen_dict

        ####### pij, degeneracy info for each orientation type observed
        verbosePrint('\t'*5, '--ij--')
        pij_degen_dict = {}
        for donors, count in Dcount.items():
            verbosePrint('\t'*5, donors, count)
            donor_constituents = Counter([str(dc) for dc in donors])
            degeneracy = 1
            sum_weight = 0
            for dc, count2 in donor_constituents.items():
                if count2 == N_H and N_H != 1:
                    degeneracy = 0
                    for num in range(shellCount[dc]-(N_H-1), 0, -1):
                        degeneracy += num
                if N_H == 1:
                    degeneracy = shellCount[dc]
                if count2 != N_H and N_H != 1:
                    degeneracy *= shellCount[dc]

            if degeneracy == 0:
                degeneracy = 1
            pij = (float(count) / float(num_orients)) / float(degeneracy)
            pij_degen_dict[donors] = [pij, degeneracy, sum_weight]

        verbosePrint('\t'*5, 'sum pair counts:', num_orients)



        solute_pijs = 0
        largest_pij = 0
        for donors, pij in pij_degen_dict.items():
            verbosePrint('\t'*6, 'pij', donors, round(pij[0], 4), 
                    'degen:', pij[1], 
                    'pij_d:', round((pij[0] * pij[1]), 4))
            solute_pij, solvent_pij = False, False
            for d in donors:
                try:
                    d = int(d)
                    if d == 0:
                        solute_pij = True
                    else:
                        continue
                except ValueError:
                    if d in ['X', 'H']:
                        solvent_pij = True
                    else:
                        solute_pij = True
            if solute_pij == True:
                solute_pijs += pij[0] * pij[1]

            if pij[0] * pij[1] > largest_pij:
                largest_pij = pij[0] * pij[1]


        ####### calculate number of observed orientations
        pij_max = max(map(lambda x: x[0], pij_degen_dict.values()))
        verbosePrint('\t'*6, 'pij_max', round(pij_max, 5))

        por_list = []
        omega_obs = 0
        for donors, pij in pij_degen_dict.items():
            om = (float(pij[0]) / float(pij_max)) * pij[1]
            por_list.append((donors, om))
            omega_obs += om


        if AD == 'D':
            soluteD_pi = solute_pis
        if AD == 'A':
            soluteA_pi = solute_pis


        verbosePrint('\t'*6, 'solute_donor', solute_donor)
        verbosePrint('\t'*6, 'Nc, Nc_solvent_only', Nc, Nc_solvent_only)
        verbosePrint('\t'*6, 'solute_pijs', round(solute_pijs, 3))
        verbosePrint('\t'*6, 'solute_pis', round(solute_pis, 3))
        verbosePrint('\t'*6, 'largest_pij', round(largest_pij, 3))
        verbosePrint('\t'*6, 'largest_pi', round(largest_pi, 3))




    S = 0
    ppD_ppA_dict = {}
    pbias_ave = 0
    Nc_eff = 0
    if Nc != 0:
        pbias_list = []
        for i, N_i in shellCount.items():
            pD_i, pA_i = 0, 0
            pD_count, pA_count, pbias = 0, 0, 0
            if i in pA_degen_dict:
                pA_i = pA_degen_dict[i][0] * pA_degen_dict[i][1]
                pA_count = pA_degen_dict[i][2]

            if i in pD_degen_dict:
                pD_i = pD_degen_dict[i][0] * pD_degen_dict[i][1]
                pD_count = pD_degen_dict[i][2]

            verbosePrint('\t'*5, 'i', i, 'N_i', N_i, 
                    'pA_count', round(pA_count, 5), 
                    'pD_count', round(pD_count, 5))
            if pD_count != 0 and pA_count != 0: 
                ppD_i = pD_i / float(pD_i + pA_i)
                ppA_i = pA_i / float(pD_i + pA_i)
                N_i_eff = ppD_i * ppA_i / float(0.25) * N_i
                Nc_eff += N_i_eff
                pbias = ppD_i * ppA_i
                for x in range(0, N_i):
                    pbias_list.append(pbias)
                verbosePrint('\t'*6, 'i', i, 'N_i', N_i, 
                        'ppA_i', round(ppA_i, 5), 
                        'ppD_i', round(ppD_i, 5), 
                        'N_i_eff', round(N_i_eff, 5), 
                        'pbias_i', round(pbias, 5))

                ppD_ppA_dict[i] = [ppD_i, ppA_i]
            else:
                for x in range(0, N_i):
                    pbias_list.append(pbias)
                if pD_count == 0 and pA_count != 0:
                    ppD_ppA_dict[i] = [0, 1]
                if pD_count != 0 and pA_count == 0:
                    ppD_ppA_dict[i] = [1, 0]
                if pD_count == 0 and pA_count == 0:
                    ppD_ppA_dict[i] = [0, 0]



        verbosePrint('\t'*5, 'Nc_eff', round(Nc_eff, 5))

        pbias_ave = 0
        if len(pbias_list) != 0:
            pbias_ave = sum(pbias_list) / float(len(pbias_list))

        verbosePrint('\t'*5, 'pbias_ave', round(pbias_ave, 5))
        if Nc_eff != 0:
            S = np.log((Nc_eff) ** 
                    (3 / float(2)) * 
                    np.pi ** 0.5 * pbias_ave / 2)
        verbosePrint('\t'*5, 'S3 (test2)', round(S, 5))

        if S < 0:
            S = 0



    return S, pbias_ave, Nc_eff, ppD_ppA_dict







def forceUnitConversion(forceUnits):
    '''
    Amber = kcal / mol/ Ang / m^2
    Gromacs = kJ / mol / nm / m^2
    '''

    ### for amber
    force_half = 0.5 ** 2
    cal_J = 4184 ** 2
    pmol_pmolec = 6.02214086e23 ** 2
    A_m = 1e-10 ** 2
    A = 1e-10
    gmol_kg = 6.02214086e26

    #'''
    if forceUnits != 'kJ':
        ###top = force_half*cal_J*gmol_kg
        top = cal_J * gmol_kg
        bottom = pmol_pmolec * A_m
        constant = float(top) / float(bottom)
    #'''

    #'''
    if forceUnits == 'kJ':
        ### for gromacs
        nm_m = 1e-9**2
        nm = 1e-9
        kJ_J = 1000 ** 2 ## sqrd as force is sqred
        top = kJ_J * gmol_kg
        bottom = pmol_pmolec * nm_m
        constant = float(top) / float(bottom)
        #constant = 1000
    #'''

    return constant






def SvibCalculations(totFrames, Aclass, temperature, forceUnits, verbosePrint):
    '''
    Get rotated forces^2 and torques^2 matrices from allMoleculeList and
    group into types.
    Then for each molecule type (Nc and extended RAD shell), calc S_vib.

    Units:
    force: Lammps = Kcal/mole-Angstrom, Amber = kcal/mol-Angstrom, 
        SI = kg.m/s^2 = N
    energy: Lammps = Kcal/mole, Amber = Kcal/mole, SI = kg.m^2/s^2 = J(/mol)
    covert Kcal/mole to KJ/mole by * 4.184

    cal to mol = * 4.184

    240.663461 = (1000/6.022E+23)/(1.38E-23*0.5) to get T from KE (kJ/mol)
    '''


    constant = forceUnitConversion(forceUnits)

    for nearest, key2 in Aclass.RADshell_dict.items():
        verbosePrint('\nnearest', nearest)
        for assigned, key3 in key2.items():
            verbosePrint('\tassigned', assigned)

            for RADshell_num, values in key3.items():
                RADshell_num = float(RADshell_num)
                verbosePrint('\n\t\tRADshell_num', RADshell_num, 
                        'count', values[5])
                forces = values[0]
                torques = values[1]
                pe = values[2]
                ke = values[3]
                Temp = values[4]
                count = values[5]
                weight = values[6]
                weightNorm = weight / float(totFrames)
                countNorm = count / float(totFrames)
                zpe = 0


                Strans_qm, Srot_qm, pe_mean, ke_mean, zpe, T = \
                        0, 0, 0, 0, 0, None

                ###setting temp here
                T = temperature 

                try:
                    force_sqrd_SI = np.multiply(forces, constant)
                    Strans_qm, zpe_trans, eigenvalues = Svib_calc(
                            force_sqrd_SI, T, 
                            cov=False, out=False, outputEigenValues=True)
                    zpe = zpe_trans
                    verbosePrint(
                            '\n\t\t\tave forces:\n\t\t\t%s'\
                                    '\n\t\t\t%s\n\t\t\t%s' 
                            % (force_sqrd_SI[0], force_sqrd_SI[1], 
                                force_sqrd_SI[2]), 
                            '\n\t\t\tStrans', Strans_qm)

                    verbosePrint('\n\t\t\teigenvalues:\n\t\t\t%s' % 
                            (eigenvalues))
                        

                    torque_sqrd_SI = np.multiply(torques, constant)
                    Srot_qm, zpe_rot, eigenvalues = Svib_calc(
                            torque_sqrd_SI, T, 
                            cov=False, out=False, outputEigenValues=True)
                    zpe += zpe_rot
                    verbosePrint(
                            '\n\t\t\tave torques:\n\t\t\t%s'\
                                    '\n\t\t\t%s\n\t\t\t%s'
                            % (torque_sqrd_SI[0], torque_sqrd_SI[1], 
                                torque_sqrd_SI[2]),
                            '\n\t\t\tSrot', Srot_qm)

                    verbosePrint('\n\t\t\teigenvalues:\n\t\t\t%s' % 
                            (eigenvalues))


                except (TypeError, ValueError): #force is a zero
                    pass


                pe = pe * 4.184 #kcal to kJ
                verbosePrint('\n\t\t\tave pe: ', pe)
                ke = ke * 4.184
                verbosePrint('\n\t\t\tave ke: ', ke)


                if Strans_qm != None:
                    Strans_qm = Strans_qm * 8.314
                if Strans_qm == None:
                    Strans_qm = 0
                if Srot_qm != None:
                    Srot_qm = Srot_qm * 8.314
                if Srot_qm == None:
                    Srot_qm = 0


                Aclass.allVariables_dict[nearest][assigned]\
                        [RADshell_num]['Strans'] = \
                        Strans_qm, count
                Aclass.allVariables_dict[nearest][assigned]\
                        [RADshell_num]['Srot'] = Srot_qm, count
                if pe != 0 and ke != 0:
                    Aclass.allVariables_dict[nearest][assigned]\
                            [RADshell_num]['PE'] = pe, count
                    Aclass.allVariables_dict[nearest][assigned]\
                            [RADshell_num]['KE'] = ke, count
                Aclass.allVariables_dict[nearest][assigned]\
                        [RADshell_num]['count'] = \
                        countNorm, count





def Svib_calc(matrix_SI, T, cov, out, *args, **kwargs):
    '''
    Calculate Svib from matrix with SI units.
    '''

    outputEigenValues = kwargs.get('outputEigenValues', False)
    eigenvalues = kwargs.get('eigenvalues', False)

    kB = 1.38064852e-23
    #if T == None:
    #T = 298 #over-write inputted T for now
    T = float(T) ##uses input T
    h = 1.0545718e-34 # value of hbar
    #h = 6.62607004e-34
    c = 30000000000
    NA = 6.022e+23
    #print(T)


    #verbosePrint(matrix_SI)

    if eigenvalues == False:
        if cov == True:
            eigenvalues, eigenvectors = LA.eig(matrix_SI)
        if cov == False:
            eigenvalues = matrix_SI.diagonal() #do this for diagonal only
                #CHECK

    
    if out == True:
        if cov == True:
            print('\n\t\t\teigenvectors:\n\t\t\t%s\n\t\t\t%s\n\t\t\t%s' 
                    % (eigenvectors[0], eigenvectors[1], 
                        eigenvectors[2]), 
                    '\n\t\t\teigenvalues:\n\t\t\t%s' % (eigenvalues))
        if cov == False:
            print('\n\t\t\teigenvalues:\n\t\t\t%s' % (eigenvalues))


    frequencies = []
    fraction = []
    sums = []
    ZPE = []
    for value in eigenvalues:
        w = value/(kB*T)
        #print(f"w = {w}")
        w = w ** 0.5
        frequencies.append(w)
        zpe = 0.5 * w * h * NA
        ZPE.append(zpe)

    for freq in frequencies:
        #this is for qm
        interim = (h*freq)/(kB*T)
        fraction.append(interim)

    for value in fraction:
        #verbosePrint(value)
        if value != 0:
            try:
                mid = (math.exp(value)-1)
                mid2 = (math.log(1-math.exp(-value)))
                mid3 = value/mid-mid2
                sums.append(mid3)
            except (ValueError, TypeError, OverflowError):
                sums.append(0)
        else:
            sums.append(0)

    S_qm = sum(sums)
    ZPE_qm = sum(ZPE)
    #Svib_qm = NA * kB * S_qm
    Svib_qm = S_qm


    if outputEigenValues == True:
        return Svib_qm, ZPE_qm, eigenvalues
    else:
        return Svib_qm, ZPE_qm
    




def WMSvibCalculations(Aclass, totFrames, temperature, 
        forceUnits, verbosePrint):
    '''
    Get forces and torques for molecule and UA levels with
    shells as x
    '''
   

    for nearest, resname_Ndih_key in \
            sorted(list(Aclass.WM_FT_shell_dict.items())):
        if 'Any' in nearest:
            verbosePrint('\n\n\n***** NEAREST: %s *****' % (nearest))
            for resname_Ndih, shellNum_key in \
                    sorted(list(resname_Ndih_key.items())):
                resname = resname_Ndih[0]
                for RADshellDist, values in \
                        sorted(list(shellNum_key.items())):
                    verbosePrint('\n\n\n***** SHELL_NUM or DIST: %s *****' % 
                            (RADshellDist))
                    RADshellDist = float(RADshellDist)
                    count = values[7]
                    #countNorm = count / float(totFrames)
                    Svib_qm, Strans_qm, Srot_qm, Strans_qm_UA, \
                            Srot_qm_UA, T_mean = process_FT(
                                    Aclass, nearest, resname_Ndih, 
                                    values, temperature, 
                                    forceUnits, verbosePrint)

                    Aclass.allVariables_dict[nearest][resname]\
                            [RADshellDist]['WM_Strans'] = \
                            Strans_qm * 8.314, count
                    Aclass.allVariables_dict[nearest][resname]\
                            [RADshellDist]['WM_Srot'] = \
                            Srot_qm * 8.314, count
                    Aclass.allVariables_dict[nearest][resname]\
                            [RADshellDist]['WM_UA_Strans'] = \
                            Strans_qm_UA * 8.314, count
                    Aclass.allVariables_dict[nearest][resname]\
                            [RADshellDist]['WM_UA_Srot'] = \
                            Srot_qm_UA * 8.314, count



def process_FT(Aclass, nearest, resname_Ndih, values, 
        temperature, forceUnits, verbosePrint):
    '''
    Get forces and torques for molecule and UA levels  - global
    '''
    

    WM_eigenvalues = []
    WM_UA_eigenvalues = []

    resname = resname_Ndih[0][0]
    Ndih = resname_Ndih[1]
    forces = values[0]
    torques = values[1]
    pe = values[2]
    ke = values[3]
    T = values[4]
    UAs_forces = values[5]
    UAs_torques = values[6]
    count = values[7]

    verbosePrint('\n\n\t***** MOLECULE LEVEL: %s *****' % (resname), 
            'count:', count)
    verbosePrint('\t\t--WM', resname)

    Strans_qm, Srot_qm, pe_mean, ke_mean, zpe, T_mean, \
            eigenvalues = FT_S(forces, torques, pe, 
            ke, forceUnits, temperature, verbosePrint, print_out=False)
    verbosePrint('\n\t\t\tTmean', T_mean, '\n')

    for ev in eigenvalues:
        e = ev + [resname]
        WM_eigenvalues.append(e)


    if type(UAs_forces) != type(int) and type(UAs_torques) != type(int):
        verbosePrint('\t\t--WM_UAs', resname)
        Strans_qm, Srot_qm, pe_mean, ke_mean, zpe, T_mean, \
                eigenvalues = FT_S(UAs_forces, UAs_torques, 
                pe, ke, forceUnits, temperature, verbosePrint, print_out=False)
        verbosePrint('\n\t\t\tTmean', T_mean, '\n')

        for ev in eigenvalues:
            e = ev + [resname]
            WM_UA_eigenvalues.append(e)


    #if T_mean != None:
        #temperature = T_mean

    verbosePrint('\t'*2, '--WM eigenvalues')
    WM_force_eigenvalues = []
    WM_torque_eigenvalues = []
    WM_eigenvalues.sort(key=lambda x: x[0])
    for ev in WM_eigenvalues:
        verbosePrint('\t'*3, ev)
        if ev[1] == 'force':
            WM_force_eigenvalues.append(ev[0])
        elif ev[1] == 'torque':
            WM_torque_eigenvalues.append(ev[0])
        else:
            continue


    WM_eigenvalues1 = sorted(list(map(lambda x: x[0], WM_eigenvalues)))
    Svib_qm, zpe_trans = Svib_calc(None, 
            temperature, cov=True, out=False, 
            eigenvalues=WM_eigenvalues1)
    verbosePrint('\n\t\t\tSvib WM', Svib_qm)

    Strans_qm, zpe_trans = Svib_calc(None, temperature, cov=True, 
            out=False, eigenvalues=WM_force_eigenvalues)
    verbosePrint('\n\t\t\tStrans WM', Strans_qm)

    Srot_qm, zpe_trans = Svib_calc(None, temperature, cov=True, 
            out=False, eigenvalues=WM_torque_eigenvalues)
    verbosePrint('\n\t\t\tSrot WM', Srot_qm)

    verbosePrint()


    verbosePrint('\t'*2, '--WM_UAs eigenvalues')
    WM_UA_eigenvalues.sort(key=lambda x: x[0])
    for ev in WM_UA_eigenvalues:
        verbosePrint('\t'*3, ev)



    verbosePrint()
    verbosePrint('\t***** WM_UA LEVEL: %s *****' % (resname))

    force_eigenvalues = []
    torque_eigenvalues = []
    for ev in WM_UA_eigenvalues:
        if ev[1] == 'force':
            force_eigenvalues.append(ev[0])
        elif ev[1] == 'torque':
            torque_eigenvalues.append(ev[0])
        else:
            continue

    torque_eigenvalues = sorted(torque_eigenvalues)
    force_eigenvalues = sorted(force_eigenvalues)

    verbosePrint('\n\t\t--All WM_UA torque eigenvalues:')
    for ev in torque_eigenvalues:
        verbosePrint('\t'*3, ev)

    verbosePrint('\n\t\t--All WM_UA force eigenvalues:')
    for ev in force_eigenvalues:
        verbosePrint('\t'*3, ev)


    force_eigenvalues = force_eigenvalues[6:] 
            # remove 6 smallest eigenvals
            #these correspond to WM level evs
    if Ndih != None: #halve Ndih largest remaining eigenvals
        verbosePrint('\n\t\tNdih: %s' % (Ndih))
        N_ev = np.array(force_eigenvalues[0:Ndih]) 
                #lowest Ndih evs are 1/2
        N_ev_halved = []
        for ev_type in N_ev:
            ev_half = float(ev_type) / float(4)
            N_ev_halved.append(ev_half)

        force_eigenvalues = N_ev_halved + force_eigenvalues[Ndih:]


    verbosePrint('\n\t\t--(-6) WM_UA torque eigenvalues:')
    for ev in torque_eigenvalues:
        verbosePrint('\t'*3, ev)

    verbosePrint('\n\t\t--(-6) WM_UA forces eigenvalues')
    for ev in force_eigenvalues:
        verbosePrint('\t'*3, ev)


    #WM_UA_eigenvalues = sorted(list(map(lambda x: x[0], 
            #WM_UA_eigenvalues)))
    Strans_qm_UA, zpe_trans_UA = Svib_calc(None, 
            temperature, cov=True, out=False,
            eigenvalues=force_eigenvalues)

    verbosePrint('\n\t\t\tStrans WM_UA', Strans_qm_UA)

    Srot_qm_UA, zpe_trans_UA = Svib_calc(None, 
            temperature, cov=True, out=False, 
            eigenvalues=torque_eigenvalues)

    verbosePrint('\n\t\t\tSrot WM_UA', Srot_qm_UA)

    #'''


    return Svib_qm, Strans_qm, Srot_qm, Strans_qm_UA, Srot_qm_UA, T_mean




def FT_S(forces, torques, pe, ke, forceUnits, temperature, 
        verbosePrint, print_out):
    '''
    '''

    constant = forceUnitConversion(forceUnits)
    forces = np.array(forces)
    torques = np.array(torques)
    pe = np.array(pe)
    ke = np.array(ke)

    Strans_qm, Srot_qm, pe_mean, ke_mean, zpe, T_mean = \
            None, None, None, None, 0, None

    ###setting temp here
    T_mean = temperature

    DOF = 0
    if type(forces) != type(int):
        for f in forces[0]:
            if f != 0:
                DOF += 1
            else:
                continue

    if type(torques) != type(int):
        for t in torques[0]:
            if t != 0:
                DOF += 1
            else:
                continue

    eigenvalues_list = []

    if type(forces) != type(int):
        force_sqrd_SI = np.multiply(forces, constant)
        Strans_qm, zpe_trans, eigenvalues = Svib_calc(force_sqrd_SI, 
                T_mean, cov=True, out=False, outputEigenValues=True)
        zpe = zpe_trans
        if print_out == True:
            verbosePrint('\n\t\t\tave forces:')
            for FF in force_sqrd_SI:
                verbosePrint('\n\t\t\t%s' % (FF))

        verbosePrint('\n\t\t\tStrans', Strans_qm)

        for ev in eigenvalues:
            eigenvalues_list.append([ev, 'force'])


    if type(torques) != type(int):
        torque_sqrd_SI = np.multiply(torques, constant)
        Srot_qm, zpe_rot, eigenvalues = Svib_calc(torque_sqrd_SI, 
                T_mean, cov=True, out=False, outputEigenValues=True)
        zpe += zpe_rot
        if print_out == True:
            verbosePrint('\n\t\t\tave torques:')
            for TT in torque_sqrd_SI:
                verbosePrint('\n\t\t\t%s' % (TT))
        verbosePrint('\n\t\t\tSrot', Srot_qm)

        for ev in eigenvalues:
            eigenvalues_list.append([ev, 'torque'])



    pe_mean = np.mean(pe) * 4.184
    verbosePrint('\n\t\t\tave pe: ', pe_mean)

    ke_mean = np.mean(ke) * 4.184
    verbosePrint('\n\t\t\tave ke: ', ke_mean)
    if DOF != 0:
        T_mean = float(ke_mean) / float(DOF) * 240.663461
    else:
        T_mean = float(ke_mean) * 240.663461


    return Strans_qm, Srot_qm, pe_mean, ke_mean, zpe, \
            T_mean, eigenvalues_list






def process_dihedrals(Aclass, verbosePrint):
    '''
    '''

    if len(Aclass.adaptive_dih_dict) != 0:

        ###adaptive dihedrals
        verbosePrint('\nAdaptive Dihedrals p ln p\n')
        for nearest, assigned_key in \
                sorted(list(Aclass.adaptive_dih_dict.items())):
            for assigned, shell_num_key in \
                    sorted(list(assigned_key.items())):
                #assigned_count = 0
                for shell_num, dih_atoms_key in \
                        sorted(list(shell_num_key.items())):
                    verbosePrint(nearest, assigned, shell_num)
                    sum_S = 0
                    shell_count = 0
                    dih_count = 0
                    for dih_atoms, phi_key in \
                            sorted(list(dih_atoms_key.items())):
                        verbosePrint(dih_atoms)
                        dih_count += 1
                        phi_count_list = []
                        max_list = []
                        for phi, count in sorted(list(phi_key.items())):
                            phi_count_list.append([phi, count])
                            shell_count += (count / 
                                    float(len(dih_atoms_key.keys())))
                        N = len(phi_count_list)-1
                        #verbosePrint(phi_count_list)
                        same_dict = nested_dict()
                        for x in range(0, N+1):
                            if phi_count_list[x][1] != 0:
                                if x == 0:
                                    if phi_count_list[x+1][1] <= \
                                            phi_count_list[x][1] and \
                                            phi_count_list[N][1] <= \
                                            phi_count_list[x][1]:
                                        max_list.append(phi_count_list[x])
                                    else:
                                        continue
                                elif x == N:
                                    if phi_count_list[x-1][1] <= \
                                            phi_count_list[x][1] and \
                                            phi_count_list[0][1] <= \
                                            phi_count_list[x][1]:
                                        max_list.append(phi_count_list[x])
                                    else:
                                        continue
                                else:
                                    if phi_count_list[x-1][1] <= \
                                            phi_count_list[x][1] and \
                                            phi_count_list[x+1][1] <= \
                                            phi_count_list[x][1]:
                                        max_list.append(phi_count_list[x])
                                    else:
                                        continue
                            else:
                                continue


                        refined_max_list = []
                        excluded = []

                        for max_count in max_list:
                            for max_count2 in max_list:
                                if max_count != max_count2 and \
                                        len(max_list) > 1 and \
                                        max_count not in refined_max_list:
                                    diff = max_count[0] - max_count2[0]

                                    normDeg = diff % 360
                                    diff2 = min(360-normDeg, normDeg)
                                    #if less than 30 degrees difference 
                                    #then only pick the largest degree to be 
                                    #binned the other gets excluded
                                    if diff2 < 60:
                                        if max_count[1] == max_count2[1]:
                                            if max_count[0] > max_count2[0] \
                                                    and \
                                                    max_count not in \
                                                    refined_max_list:
                                                refined_max_list.append(
                                                        max_count)
                                                excluded.append(max_count2)
                                            else:
                                                continue
                                        elif max_count[1] > max_count2[1]:
                                            if max_count not in \
                                                    refined_max_list:
                                                refined_max_list.append(
                                                        max_count)
                                            else:
                                                continue
                                        else:
                                            continue
                                elif max_count == max_count2 and \
                                        len(max_list) == 1 and \
                                        max_count not in refined_max_list:
                                    refined_max_list.append(max_count)
                                else:
                                    continue

                        for max_count in max_list:
                            for max_count2 in max_list:
                                if max_count != max_count2 and \
                                        len(max_list) > 1 and \
                                        max_count not in refined_max_list:
                                    diff = max_count[0] - max_count2[0]
                                    normDeg = diff % 360
                                    diff2 = min(360-normDeg, normDeg)
                                    if diff2 >= 60 and max_count not in \
                                            refined_max_list and \
                                            max_count not in excluded:
                                        refined_max_list.append(max_count)

                        refined_max_dict = {}
                        for max_count in refined_max_list:
                            refined_max_dict[max_count[0]] = max_count[1]
                        max_phis = list(refined_max_dict.keys())
                        for phi, count in phi_count_list:
                            if phi not in refined_max_dict and count != 0:
                                smallest_diff = 999
                                closest_phi = None
                                for maxPhi in max_phis:
                                    diff = phi - maxPhi
                                    normDeg = diff % 360
                                    diff2 = min(360-normDeg, normDeg)
                                    if diff2 < smallest_diff:
                                        smallest_diff = diff2
                                        closest_phi = maxPhi
                                refined_max_dict[closest_phi] += count

                        ## p ln p calcs finally!
                        sum_count = sum(count 
                                for phi, count in refined_max_dict.items())
                        topo_S = 0
                        for phi, count in refined_max_dict.items():
                            p = count / float(sum_count)
                            S = -p * np.log(p) * 8.314
                            topo_S += S
                            verbosePrint('\t', phi, count, p, S)
                        verbosePrint('\t\t', topo_S)
                        verbosePrint()
                        sum_S += topo_S
                    
                    Aclass.allVariables_dict[nearest][(assigned[0], \
                            assigned[3], assigned[4])]\
                            [float(shell_num)]['conf_AE'] = \
                            sum_S, int(round(shell_count, 0))
                    verbosePrint('\nS_adaptive', sum_S, 'N_dih:', dih_count, 
                            assigned)
                    verbosePrint()






def saveAllVariables(num_frames, Aclass, level, name, solvent, 
        waterTuple, verbosePrint):
    '''
    '''

    frames = ''
    if num_frames != None:
        frames = num_frames
    solventDataName = f'solventVariables{frames}{name}_{level}.csv'
    #data = open('solventVariables%s%s_%s.csv' % (frames, name, level), 'w')
    # data.write(
    #         '\n'.join(['nearest,assigned,shell_num,variable,value,count']) 
    #         + '\n')
    Utils.printOut(solventDataName,'nearest,assigned,shell_num,variable,value,count')
    solventData = pd.DataFrame(columns=['nearest','assigned', 'shell_num', 'variable', 'value', 'count'])

    soluteDataName = f'soluteVariables{frames}{name}_{level}.csv'
    # data2 = open('soluteVariables%s%s_%s.csv' % 
    #         (frames, name, level), 'w')
    # data2.write('\n'.join(['resName,variable,value,count']) 
    #         + '\n')
    Utils.printOut(soluteDataName,'resName,variable,value,count')

    soluteData = pd.DataFrame(columns=['resName', 'variable', 'value', 'count'])


    if level == 'res_atomLevel':
        reduced_dict = nested_dict()
        for nearest, assigned_key in \
                sorted(list(Aclass.allVariables_dict.items())):
            nearest = nearest.split('_')[0]
            for assigned, shellNum_key in sorted(list(assigned_key.items())):
                assigned = assigned[0].split('_')[0]
                if assigned in waterTuple:
                    for shellNum, variable_key in \
                            sorted(list(shellNum_key.items())):
                        tot_count = 0
                        for nearest2, assigned2_key in \
                                Aclass.allVariables_dict.items():
                            for assigned2, shellNum2_key in \
                                    assigned2_key.items():
                                for shellNum2, variable2_key in \
                                        shellNum2_key.items():
                                    for variable2, count in \
                                            variable2_key.items():
                                        nearest3 = nearest2.split('_')[0]
                                        assigned3 = \
                                                assigned2[0].split('_')[0]
                                        if nearest3 == nearest \
                                                and assigned3 == assigned \
                                                and shellNum2 == shellNum \
                                                and variable2 == 'count':
                                            tot_count += count[1]
                                        else:
                                            continue
                        for variable, value in variable_key.items():
                            if variable not in \
                                    reduced_dict[nearest][assigned]\
                                    [shellNum] and value[0] != None:
                                reduced_dict[nearest][assigned][shellNum]\
                                        [variable] = [0, 0]
                            reduced_dict[nearest][assigned][shellNum]\
                                    [variable][0] += \
                                    (value[0] * value[1] / 
                                            float(tot_count))
                            reduced_dict[nearest][assigned][shellNum]\
                                    [variable][1] += \
                                    int(round(value[1], 0))
                else:
                    continue


        for nearest, assigned_key in \
                sorted(list(reduced_dict.items())):
            for assigned, shellNum_key in sorted(list(assigned_key.items())):
                for shellNum, variable_key in \
                        sorted(list(shellNum_key.items())):
                    for variable, value in variable_key.items():
                        if value[0] != None and \
                                assigned in solvent and shellNum == 1:
                            # data.write('\n'.join(['%s,%s,%s,%s,%s,%s' % 
                            #         (nearest, 
                            #         assigned, shellNum, variable, value[0], 
                            #         value[1])]) + '\n')
                            Utils.printOut(solventDataName, f"{nearest},{assigned},{shellNum},{variable},{value[0]},{value[1]}")
                            newRowSolvent = pd.DataFrame({'nearest': nearest,
                                            'assigned':assigned,
                                            'shell_num': shellNum,
                                            'variable': variable,
                                            'value': value[0],
                                            'count': value[1]}, index=[0])
                            solventData = pd.concat([solventData, newRowSolvent], ignore_index=True)
                        else:
                            continue

        ###for solute only
        PE_KE_dict = nested_dict()
        for nearest, assigned_key in \
                sorted(list(Aclass.allVariables_dict.items())):
            for assigned, shellNum_key in sorted(list(assigned_key.items())):
                assigned = list(assigned)
                if '_' in assigned[0]:
                    assigned[0] = assigned[0].split('_')[0]
                for shellNum, variable_key in \
                        sorted(list(shellNum_key.items())): 
                    for variable, value in variable_key.items():
                        if assigned[0] not in solvent and shellNum != 0:
                            if variable in ['PE', 'KE'] and \
                                    value[0] != None:
                                new_var = 'WM_%s' % (variable)
                                if new_var not in PE_KE_dict[nearest]\
                                        [assigned[0]]:
                                    PE_KE_dict[nearest][assigned[0]]\
                                            [new_var] = [0, 0]
                                PE_KE_dict[nearest][assigned[0]]\
                                        [new_var][0] += value[0]
                                PE_KE_dict[nearest][assigned[0]]\
                                        [new_var][1] = value[1]

                            if 'WM_' in variable or variable == 'conf_AE':
                                # data2.write('\n'.join(['%s,%s,%s,%s' % 
                                #         (assigned[0], 
                                #         variable, value[0], 
                                #         int(round(value[1], 0)))]) + '\n')
                                Utils.printOut(soluteDataName,f"{assigned[0]},{variable},{value[0]},{int(round(value[1], 0))}")
                                newRowSolute = pd.DataFrame({'resName': assigned[0],
                                            'variable':variable,
                                            'value': value[0],
                                            'count': int(round(value[1], 0))}, index=[0])
                                soluteData = pd.concat([soluteData, newRowSolute], ignore_index=True)            
                        else:
                            continue

       
        for nearest, assigned_key in \
                sorted(list(PE_KE_dict.items())):
            for assigned, variable_key in sorted(list(assigned_key.items())):
                for variable, value in sorted(list(variable_key.items())):
                    # data2.write('\n'.join(['%s,%s,%s,%s' % 
                    #         (assigned, 
                    #         variable, value[0], 
                    #         value[1])]) + '\n')
                    Utils.printOut(soluteDataName,f"{assigned[0]},{variable},{value[0]},{value[1]}")
                    newRowSolute = pd.DataFrame({'resName': assigned,
                                    'variable':variable,
                                    'value': value[0],
                                    'count': value[1]}, index=[0])
                    soluteData = pd.concat([soluteData, newRowSolute], ignore_index=True)
        #'''

    else:
        for nearest, assigned_key in \
                sorted(list(Aclass.allVariables_dict.items())):
            for assigned, shellNum_key in sorted(list(assigned_key.items())):
                for shellNum, variable_key in \
                        sorted(list(shellNum_key.items())): 
                    for variable, value in variable_key.items():
                        if assigned[0] in solvent and shellNum == 1 and \
                                value[0] != None:
                            # data.write('\n'.join(['%s,%s,%s,%s,%s,%s' % 
                            #         (nearest, 
                            #         assigned[0], shellNum, 
                            #         variable, value[0], 
                            #         int(round(value[1], 0)))]) + '\n')
                            Utils.printOut(solventDataName, f"{nearest},{assigned[0]},{shellNum},{variable},{value[0]},{int(round(value[1], 0))}")
                            newRowSolvent = pd.DataFrame({'nearest': nearest,
                                            'assigned':assigned[0],
                                            'shell_num': shellNum,
                                            'variable': variable,
                                            'value': value[0],
                                            'count': int(round(value[1], 0))}, index=[0])
                            solventData = pd.concat([solventData, newRowSolvent], ignore_index=True)
                        else:
                            continue


        #'''
        ###for solute only
        PE_KE_dict = nested_dict()
        for nearest, assigned_key in \
                sorted(list(Aclass.allVariables_dict.items())):
            for assigned, shellNum_key in sorted(list(assigned_key.items())):
                for shellNum, variable_key in \
                        sorted(list(shellNum_key.items())): 
                    for variable, value in variable_key.items():
                        if assigned[0] not in solvent and shellNum != 0:
                            if variable in ['PE', 'KE'] and \
                                    value[0] != None:
                                new_var = 'WM_%s' % (variable)
                                if new_var not in PE_KE_dict[nearest]\
                                        [assigned[0]]:
                                    PE_KE_dict[nearest][assigned[0]]\
                                            [new_var] = [0, 0]
                                PE_KE_dict[nearest][assigned[0]]\
                                        [new_var][0] += value[0]
                                PE_KE_dict[nearest][assigned[0]]\
                                        [new_var][1] = value[1]

                            if 'WM_' in variable or variable == 'conf_AE':
                                # data2.write('\n'.join(['%s,%s,%s,%s' % 
                                #         (assigned[0], 
                                #         variable, value[0], 
                                #         int(round(value[1], 0)))]) + '\n')
                                Utils.printOut(soluteDataName,f"{assigned[0]},{variable},{value[0]},{int(round(value[1], 0))}")
                                newRowSolute = pd.DataFrame({'resName': assigned[0],
                                                'variable':variable,
                                                'value': value[0],
                                                'count': int(round(value[1], 0))}, index=[0])
                                soluteData = pd.concat([soluteData, newRowSolute], ignore_index=True)

                            
                        else:
                            continue

        for nearest, assigned_key in \
                sorted(list(PE_KE_dict.items())):
            for assigned, variable_key in sorted(list(assigned_key.items())):
                for variable, value in sorted(list(variable_key.items())):
                    # data2.write('\n'.join(['%s,%s,%s,%s' % 
                    #         (assigned, 
                    #         variable, value[0], 
                    #         value[1])]) + '\n')
                    Utils.printOut(soluteDataName,f"{assigned},{variable},{value[0]},{value[1]}")
                    newRowSolute = pd.DataFrame({'resName': assigned[0],
                                    'variable':variable,
                                    'value': value[0],
                                    'count': int(round(value[1], 0))}, index=[0])
                    soluteData = pd.concat([soluteData, newRowSolute], ignore_index=True)
        #'''
    

    # data.close()
    # data2.close()
    return (solventData, soluteData)



import os, sys
import MDAnalysis as mda
import datetime as dt
from CodeEntropy.ClassCollection.PoseidonClass import Poseidon
import pandas as pd
import pytest
from numpy import array, testing

@pytest.fixture(scope='session')
def wd():
    wd = os.path.dirname(os.path.abspath(__file__))
    return wd
@pytest.fixture(scope='session')
def poseidon_object(wd):
    topo_file = os.path.join(wd,"data/poseidon_example.prmtop")
    traj_file = os.path.join(wd,"data/poseidon_example.trr")
    u = mda.Universe(topo_file, traj_file)
    poseidon_object = Poseidon(container=u, start=2, end=12)
    return poseidon_object

parser_ref_index = [
    (0, 9),
    (1, 916), 
    (2, 4584), 
    (3, 2093), 
    (4, 8641), 
    (5, 5221)
]
@pytest.fixture(scope='session')
def parser_ref():
    parser_ref = [
        [2,
        'CZ',
        21,
        'ARG',
        2,
        [3, 0],
        [6, 8, 10, 13, 16, 19, 21, 22, 25, 28, 29],
        ['O', 'WAT', 240, 3.307],
        1,
        (['0', 'X'], [], []),
        None,
        None,
        None,
        array([[ 14.508, -29.179, -13.636],
                [-29.179,  58.689,  27.426],
                [-13.636,  27.426,  12.816]]),
        array([[0., 0., 0.],
                [0., 0., 0.],
                [0., 0., 0.]]),
        None,
        None,
        None,
        [['CD', 2, 'ARG'], ['O', 240, 'WAT']]],
        [3,
        'CH3',
        1,
        'ACE',
        1,
        [1, 3],
        [1, 4, 5],
        ['N', 'ARG', 2, 2.428],
        1,
        (['0', '1', 'WAT_O', 'WAT_O', 'WAT_O', 'WAT_O', 'WAT_O'], [], []),
        array([[ 0.258, -0.   ,  0.206],
                [-0.   ,  0.   , -0.   ],
                [ 0.206, -0.   ,  0.165]]),
        array([[ 4.133,  1.435, -6.39 ],
                [ 1.435,  0.498, -2.219],
                [-6.39 , -2.219,  9.882]]),
        None,
        array([[ 2.000e-03,  6.400e-02, -1.420e-01],
                [ 6.400e-02,  1.845e+00, -4.090e+00],
                [-1.420e-01, -4.090e+00,  9.066e+00]]),
        array([[ 0.657,  5.304,  4.153],
                [ 5.304, 42.808, 33.517],
                [ 4.153, 33.517, 26.243]]),
        [],
        array([[  0.,   0.,  -1.,   0.,  -0.,   1.,  -0.,  -0.,  -0.],
                [  0.,   7., -16.,  12.,  -2.,  32.,  -6.,  -6.,  -8.],
                [ -1., -16.,  36., -27.,   4., -70.,  14.,  13.,  17.],
                [  0.,  12., -27.,  20.,  -3.,  51., -10.,  -9., -13.],
                [ -0.,  -2.,   4.,  -3.,   0.,  -7.,   1.,   1.,   2.],
                [  1.,  32., -70.,  51.,  -7., 134., -26., -24., -33.],
                [ -0.,  -6.,  14., -10.,   1., -26.,   5.,   5.,   6.],
                [ -0.,  -6.,  13.,  -9.,   1., -24.,   5.,   4.,   6.],
                [ -0.,  -8.,  17., -13.,   2., -33.,   6.,   6.,   8.]]),
        array([[ 0.657,  5.304,  4.153,  0.   ,  0.   ,  0.   ,  0.   ,  0.   ,
                0.   ],
                [ 5.304, 42.808, 33.517,  0.   ,  0.   ,  0.   ,  0.   ,  0.   ,
                0.   ],
                [ 4.153, 33.517, 26.243,  0.   ,  0.   ,  0.   ,  0.   ,  0.   ,
                0.   ],
                [ 0.   ,  0.   ,  0.   ,  0.   ,  0.   ,  0.   ,  0.   ,  0.   ,
                0.   ],
                [ 0.   ,  0.   ,  0.   ,  0.   ,  0.   ,  0.   ,  0.   ,  0.   ,
                0.   ],
                [ 0.   ,  0.   ,  0.   ,  0.   ,  0.   ,  0.   ,  0.   ,  0.   ,
                0.   ],
                [ 0.   ,  0.   ,  0.   ,  0.   ,  0.   ,  0.   ,  0.   ,  0.   ,
                0.   ],
                [ 0.   ,  0.   ,  0.   ,  0.   ,  0.   ,  0.   ,  0.   ,  0.   ,
                0.   ],
                [ 0.   ,  0.   ,  0.   ,  0.   ,  0.   ,  0.   ,  0.   ,  0.   ,
                0.   ]]),
        [['O', 1, 'ACE'],
        ['N', 2, 'ARG'],
        ['O', 817, 'WAT'],
        ['O', 61, 'WAT'],
        ['O', 539, 'WAT'],
        ['O', 341, 'WAT'],
        ['O', 315, 'WAT']]],
        [7,
        'CA',
        8,
        'ARG',
        2,
        [3, 1],
        [6, 8, 10, 13, 16, 19, 21, 22, 25, 28, 29],
        ['C', 'ACE', 1, 2.484],
        1,
        (['0', 'NME_N', 'WAT_O', 'X', 'X'], [], []),
        None,
        None,
        None,
        array([[ 5.057, 16.631, 18.57 ],
                [16.631, 54.691, 61.068],
                [18.57 , 61.068, 68.188]]),
        array([[17.433, 14.215, 33.262],
                [14.215, 11.59 , 27.121],
                [33.262, 27.121, 63.463]]),
        None,
        None,
        None,
        [['C', 1, 'ACE'],
        ['O', 2, 'ARG'],
        ['N', 3, 'NME'],
        ['CG', 2, 'ARG'],
        ['O', 422, 'WAT']]],
        [4,
        'O',
        771,
        'WAT',
        249,
        [0, 2],
        [771],
        [],
        2,
        ([], [], []),
        array([[ 0.009,  0.027, -0.004],
                [ 0.027,  0.076, -0.01 ],
                [-0.004, -0.01 ,  0.001]]),
        array([[ 0.459,  0.477, -0.352],
                [ 0.477,  0.496, -0.365],
                [-0.352, -0.365,  0.269]]),
        None,
        array([[ 0.009,  0.027, -0.004],
                [ 0.027,  0.076, -0.01 ],
                [-0.004, -0.01 ,  0.001]]),
        array([[ 0.459,  0.477, -0.352],
                [ 0.477,  0.496, -0.365],
                [-0.352, -0.365,  0.269]]),
        None,
        array([[ 0.009,  0.027, -0.004],
                [ 0.027,  0.076, -0.01 ],
                [-0.004, -0.01 ,  0.001]]),
        array([[ 0.459,  0.477, -0.352],
                [ 0.477,  0.496, -0.365],
                [-0.352, -0.365,  0.269]]),
        []],
        [11,
        'O',
        1179,
        'WAT',
        385,
        [0, 2],
        [1179],
        [],
        2,
        ([], [], []),
        array([[0.   , 0.   , 0.   ],
                [0.   , 0.351, 0.102],
                [0.   , 0.102, 0.03 ]]),
        array([[ 0.77 ,  0.908, -1.873],
                [ 0.908,  1.07 , -2.208],
                [-1.873, -2.208,  4.557]]),
        None,
        array([[0.   , 0.   , 0.   ],
                [0.   , 0.351, 0.102],
                [0.   , 0.102, 0.03 ]]),
        array([[ 0.77 ,  0.908, -1.873],
                [ 0.908,  1.07 , -2.208],
                [-1.873, -2.208,  4.557]]),
        None,
        array([[0.   , 0.   , 0.   ],
                [0.   , 0.351, 0.102],
                [0.   , 0.102, 0.03 ]]),
        array([[ 0.77 ,  0.908, -1.873],
                [ 0.908,  1.07 , -2.208],
                [-1.873, -2.208,  4.557]]),
        []],
        [7,
        'O',
        1911,
        'WAT',
        629,
        [0, 2],
        [1911],
        [],
        2,
        ([], [], []),
        array([[0.024, 0.05 , 0.223],
                [0.05 , 0.103, 0.459],
                [0.223, 0.459, 2.037]]),
        array([[ 2.0345e+01, -4.8700e-01, -1.1815e+01],
                [-4.8700e-01,  1.2000e-02,  2.8300e-01],
                [-1.1815e+01,  2.8300e-01,  6.8610e+00]]),
        None,
        array([[0.024, 0.05 , 0.223],
                [0.05 , 0.103, 0.459],
                [0.223, 0.459, 2.037]]),
        array([[ 2.0345e+01, -4.8700e-01, -1.1815e+01],
                [-4.8700e-01,  1.2000e-02,  2.8300e-01],
                [-1.1815e+01,  2.8300e-01,  6.8610e+00]]),
        None,
        array([[0.024, 0.05 , 0.223],
                [0.05 , 0.103, 0.459],
                [0.223, 0.459, 2.037]]),
        array([[ 2.0345e+01, -4.8700e-01, -1.1815e+01],
                [-4.8700e-01,  1.2000e-02,  2.8300e-01],
                [-1.1815e+01,  2.8300e-01,  6.8610e+00]]),
        []]
    ]
    return parser_ref

# the reason for testing individually is to allow identification of which value went wrong
@pytest.mark.parametrize("ref_index,test_id",
    parser_ref_index
)
def test_poseidon_parser_frame(poseidon_object, parser_ref, ref_index,test_id):
    """
    testing the return of allMoleculeList index 0 which store the frame of the atom info
    """
    assert poseidon_object.allMoleculeList[test_id][0] == parser_ref[ref_index][0]

@pytest.mark.parametrize("ref_index,test_id",
    parser_ref_index
)
def test_poseidon_parser_atom_name(poseidon_object, parser_ref, ref_index,test_id):
    """
    testing the return of allMoleculeList index 1 which store the atom name
    """
    assert poseidon_object.allMoleculeList[test_id][1] == parser_ref[ref_index][1]

@pytest.mark.parametrize("ref_index,test_id",
    parser_ref_index
)
def test_poseidon_parser_atom_num(poseidon_object, parser_ref, ref_index,test_id):
    """
    testing the return of allMoleculeList index 2 which store the atom number
    """
    assert poseidon_object.allMoleculeList[test_id][2] == parser_ref[ref_index][2]

@pytest.mark.parametrize("ref_index,test_id",
    parser_ref_index
)
def test_poseidon_parser_resname(poseidon_object, parser_ref, ref_index,test_id):
    """
    testing the return of allMoleculeList index 3 which store the resname
    """
    assert poseidon_object.allMoleculeList[test_id][3] == parser_ref[ref_index][3]

@pytest.mark.parametrize("ref_index,test_id",
    parser_ref_index
)
def test_poseidon_parser_resid(poseidon_object, parser_ref, ref_index,test_id):
    """
    testing the return of allMoleculeList index 4 which store the resid
    """
    assert poseidon_object.allMoleculeList[test_id][4] == parser_ref[ref_index][4]

@pytest.mark.parametrize("ref_index,test_id",
    parser_ref_index
)
def test_poseidon_parser_bondedUA_H(poseidon_object, parser_ref, ref_index,test_id):
    """
    testing the return of allMoleculeList index 5 which store the bondedUA_H which is a tuple of [num_bondedUAs, num_bonded Hs]
    """
    assert poseidon_object.allMoleculeList[test_id][5] == parser_ref[ref_index][5]

@pytest.mark.parametrize("ref_index,test_id",
    parser_ref_index
)
def test_poseidon_parser_molecule_atomNums(poseidon_object, parser_ref, ref_index,test_id):
    """
    testing the return of allMoleculeList index 6 which store the list of bonded UA atom nums
    """
    assert poseidon_object.allMoleculeList[test_id][6] == parser_ref[ref_index][6]

@pytest.mark.parametrize("ref_index,test_id",
    parser_ref_index
)
def test_poseidon_parser_nearestInfo(poseidon_object, parser_ref, ref_index,test_id):
    """
    testing the return of allMoleculeList index 7 which store the list of nearest non like neighbour in a tuple of (atom_name, resname, resid, distance)
    """
    assert poseidon_object.allMoleculeList[test_id][7] == parser_ref[ref_index][7]

@pytest.mark.parametrize("ref_index,test_id",
    parser_ref_index
)
def test_poseidon_parser_hydrationShellRAD_solute(poseidon_object, parser_ref, ref_index,test_id):
    """
    testing the return of allMoleculeList index 8 which store the list of nearest solute molecule centric
    """
    assert poseidon_object.allMoleculeList[test_id][8] == parser_ref[ref_index][8]

@pytest.mark.parametrize("ref_index,test_id",
    parser_ref_index
)
def test_poseidon_parser_RAD_shell_ranked(poseidon_object, parser_ref, ref_index,test_id):
    """
    testing the return of allMoleculeList index 9 which store the list of 3 tuple (RAD shell ranked, acceptors ranked, donors ranked by RAD)
    """
    assert poseidon_object.allMoleculeList[test_id][9] == parser_ref[ref_index][9]

@pytest.mark.parametrize("ref_index,test_id",
    parser_ref_index
)
def test_poseidon_parser_MweightedForces(poseidon_object, parser_ref, ref_index,test_id):
    """
    testing the return of allMoleculeList index 10 which store the MweightedForces
    """
    #numpy can't compare None type object
    if not parser_ref[ref_index][10] is None:
        testing.assert_array_almost_equal(poseidon_object.allMoleculeList[test_id][10], parser_ref[ref_index][10])
    else:
        assert poseidon_object.allMoleculeList[test_id][10] == parser_ref[ref_index][10]

@pytest.mark.parametrize("ref_index,test_id",
    parser_ref_index
)
def test_poseidon_parser_MweightedTorques(poseidon_object, parser_ref, ref_index,test_id):
    """
    testing the return of allMoleculeList index 11 which store the MweightedTorques
    """
    #numpy can't compare None type object
    if not parser_ref[ref_index][11] is None:
        testing.assert_array_almost_equal(poseidon_object.allMoleculeList[test_id][11], parser_ref[ref_index][11])
    else:
        assert poseidon_object.allMoleculeList[test_id][11] == parser_ref[ref_index][11]

@pytest.mark.parametrize("ref_index,test_id",
    parser_ref_index
)
def test_poseidon_parser_UA_PKenergy(poseidon_object, parser_ref, ref_index,test_id):
    """
    testing the return of allMoleculeList index 12 which store the UA_PKenergy should all be None as energy is out of scope
    """
    assert poseidon_object.allMoleculeList[test_id][12] == parser_ref[ref_index][12]

@pytest.mark.parametrize("ref_index,test_id",
    parser_ref_index
)
def test_poseidon_parser_UAweightedForces(poseidon_object, parser_ref, ref_index,test_id):
    """
    testing the return of allMoleculeList index 13 which store the UAweightedForces
    """
    #numpy can't compare None type object
    if not parser_ref[ref_index][13] is None:
        testing.assert_array_almost_equal(poseidon_object.allMoleculeList[test_id][13], parser_ref[ref_index][13])
    else:
        assert poseidon_object.allMoleculeList[test_id][13] == parser_ref[ref_index][13]

@pytest.mark.parametrize("ref_index,test_id",
    parser_ref_index
)
def test_poseidon_parser_UAweightedTorques(poseidon_object, parser_ref, ref_index,test_id):
    """
    testing the return of allMoleculeList index 14 which store the UAweightedTorques
    """
    #numpy can't compare None type object
    if not parser_ref[ref_index][14] is None:
        testing.assert_array_almost_equal(poseidon_object.allMoleculeList[test_id][14], parser_ref[ref_index][14])
    else:
        assert poseidon_object.allMoleculeList[test_id][14] == parser_ref[ref_index][14]

@pytest.mark.parametrize("ref_index,test_id",
    parser_ref_index
)
def test_poseidon_parser_dihedral_phi_type(poseidon_object, parser_ref, ref_index,test_id):
    """
    testing the return of allMoleculeList index 15 which store the dihedral_phi_type a list of unique dihedral angles, in degrees
    """
    assert poseidon_object.allMoleculeList[test_id][15] == parser_ref[ref_index][15]

@pytest.mark.parametrize("ref_index,test_id",
    parser_ref_index
)
def test_poseidon_parser_molecule_UA_Fs(poseidon_object, parser_ref, ref_index,test_id):
    """
    testing the return of allMoleculeList index 16 which store the molecule_UA_Fs
    """
    #numpy can't compare None type object
    if not parser_ref[ref_index][16] is None:
        testing.assert_array_almost_equal(poseidon_object.allMoleculeList[test_id][16], parser_ref[ref_index][16])
    else:
        assert poseidon_object.allMoleculeList[test_id][16] == parser_ref[ref_index][16]

@pytest.mark.parametrize("ref_index,test_id",
    parser_ref_index
)
def test_poseidon_parser_molecule_UA_Ts(poseidon_object, parser_ref, ref_index,test_id):
    """
    testing the return of allMoleculeList index 17 which store the molecule_UA_Ts
    """
    #numpy can't compare None type object
    if not parser_ref[ref_index][17] is None:
        testing.assert_array_almost_equal(poseidon_object.allMoleculeList[test_id][17], parser_ref[ref_index][17])
    else:
        assert poseidon_object.allMoleculeList[test_id][17] == parser_ref[ref_index][17]

@pytest.mark.parametrize("ref_index,test_id",
    parser_ref_index
)
def test_poseidon_parser_RAD_nAtoms(poseidon_object, parser_ref, ref_index,test_id):
    """
    testing the return of allMoleculeList index 18 which store the RAD_nAtoms
    """
    assert poseidon_object.allMoleculeList[test_id][18] == parser_ref[ref_index][18]




def test_poseidon_moleculeLevel(poseidon_object, wd):
    """
    testing result of poseidon at moleculeLevel
    """
    result = poseidon_object.run_analysis(level_list = ['moleculeLevel'], verbose=False)
    data_dir_solute = os.path.join(wd,"data/soluteVariables10.0EE_moleculeLevel.csv")
    solute_ref = pd.read_csv(data_dir_solute, na_values="nan")
    pd.testing.assert_frame_equal(solute_ref, result["moleculeLevel"]["soluteData"], check_dtype=False)
    data_dir_solvent = os.path.join(wd,"data/solventVariables10.0EE_moleculeLevel.csv")
    solvent_ref = pd.read_csv(data_dir_solvent, na_values="nan")
    pd.testing.assert_frame_equal(solvent_ref, result["moleculeLevel"]["solventData"], check_dtype=False)
    
def test_poseidon_residLevel_resname(poseidon_object, wd):
    """
    testing result of poseidon at residLevel_resname level
    """
    result = poseidon_object.run_analysis(level_list = ['residLevel_resname'], verbose=False)
    data_dir_solute = os.path.join(wd,"data/soluteVariables10.0EE_residLevel_resname.csv")
    solute_ref = pd.read_csv(data_dir_solute, na_values="nan")
    pd.testing.assert_frame_equal(solute_ref, result["residLevel_resname"]["soluteData"], check_dtype=False)
    data_dir_solvent = os.path.join(wd,"data/solventVariables10.0EE_residLevel_resname.csv")
    solvent_ref = pd.read_csv(data_dir_solvent, na_values="nan")
    pd.testing.assert_frame_equal(solvent_ref, result["residLevel_resname"]["solventData"], check_dtype=False)
    data_dir_contact = os.path.join(wd,"data/resid_contact_matrix_residLevel_resname.csv")
    contact_ref = pd.read_csv(data_dir_contact, na_values="nan")
    pd.testing.assert_frame_equal(contact_ref, result["residLevel_resname"]["contactMatrix"], check_dtype=False)

def test_poseidon_atomLevel(poseidon_object, wd):
    """
    testing result of poseidon at atomLevel
    """
    result = poseidon_object.run_analysis(level_list = ['atomLevel'], verbose=False)
    data_dir_solute = os.path.join(wd,"data/soluteVariables10.0EE_atomLevel.csv")
    solute_ref = pd.read_csv(data_dir_solute, na_values="nan")
    pd.testing.assert_frame_equal(solute_ref, result["atomLevel"]["soluteData"], check_dtype=False)
    data_dir_solvent = os.path.join(wd,"data/solventVariables10.0EE_atomLevel.csv")
    solvent_ref = pd.read_csv(data_dir_solvent, na_values="nan")
    pd.testing.assert_frame_equal(solvent_ref, result["atomLevel"]["solventData"], check_dtype=False)
    data_dir_contact = os.path.join(wd,"data/resid_contact_matrix_atomLevel.csv")
    contact_ref = pd.read_csv(data_dir_contact, na_values="nan")
    pd.testing.assert_frame_equal(contact_ref, result["atomLevel"]["contactMatrix"], check_dtype=False)

def test_poseidon_soluteContacts(poseidon_object, wd):
    """
    testing result of poseidon at soluteContacts Level
    """
    result = poseidon_object.run_analysis(level_list = ['soluteContacts'], verbose=False)
    data_dir_solute = os.path.join(wd,"data/soluteVariables10.0EE_soluteContacts.csv")
    solute_ref = pd.read_csv(data_dir_solute, na_values="nan")
    pd.testing.assert_frame_equal(solute_ref, result["soluteContacts"]["soluteData"], check_dtype=False)
    data_dir_solvent = os.path.join(wd,"data/solventVariables10.0EE_soluteContacts.csv")
    solvent_ref = pd.read_csv(data_dir_solvent, na_values="nan")
    pd.testing.assert_frame_equal(solvent_ref, result["soluteContacts"]["solventData"], check_dtype=False)


import psutil
import os

def memoryInfo(verbosePrint):
    '''
    '''
    process = psutil.Process(os.getpid())
    bytes_info = float(process.memory_info().rss)
    gb = bytes_info * 1e-9
    print('memory use:', round(gb, 3), 'GB')  # in gbytes




def weightingPopulation(weighting):
    '''
    Save a list of weightings per frame from biased simulations in the
    EEclass
    '''
    dataFile = weighting[0]
    try:
        frame_chunks = weighting[1]
    except IndexError:
        frame_chunks = 1
    weighting_list = []
    with open(dataFile) as data:
        for line in data:
            weighting_list.append(line)
    data.close()

    weighting_info = [weighting_list, frame_chunks]

    return weighting_info



#!/usr/bin/env python

import numpy as np
from itertools import combinations




def contactPopulation(Aclass, RAD_nAtoms, atom_resname, 
        atom_resid, waterTuple, weight):
    '''
    '''

    ####### RESID CONTACT MATRIX POP
    ### contacts defined using QcQn/r^2 between any atoms 
    ### (H or heavy atoms) here!
    if RAD_nAtoms != None:
        #RAD_closestRanked = []
        #nRAD_closestRanked_separated = {}
        for nInfo in RAD_nAtoms:
            nResid = nInfo[1]
            nResname = nInfo[2]
            if nResname not in waterTuple and \
                    atom_resname not in waterTuple and\
                    nResid != atom_resid:
                if (nResname, nResid) not in \
                        Aclass.resid_contact_matrix_dict\
                        [(atom_resname, atom_resid)]:
                    Aclass.resid_contact_matrix_dict\
                            [(atom_resname, atom_resid)]\
                            [(nResname, nResid)] = 0
                    Aclass.resid_contact_matrix_dict\
                            [(nResname, nResid)]\
                            [(atom_resname, atom_resid)] = 0
                Aclass.resid_contact_matrix_dict\
                        [(atom_resname, atom_resid)]\
                        [(nResname, nResid)] += weight #1

            else:
                continue


def contactPopulationUA(Aclass, RAD_nAtoms, atom_name, atom_resname, 
        atom_resid, waterTuple, weight):
    '''
    '''

    ####### RESID CONTACT MATRIX POP
    ### contacts defined using QcQn/r^2 between any atoms 
    ### (H or heavy atoms) here!
    atom_resname = atom_resname.split('_')[0]
    if RAD_nAtoms != None:
        #RAD_closestRanked = []
        #nRAD_closestRanked_separated = {}
        for nInfo in RAD_nAtoms:
            nAtom = nInfo[0]
            nResid = nInfo[1]
            nResname = nInfo[2]
            if nResname not in waterTuple and \
                    atom_resname not in waterTuple and\
                    nResid != atom_resid:
                if (nAtom, nResname, nResid) not in \
                        Aclass.resid_contact_matrix_dict\
                        [(atom_name, atom_resname, atom_resid)]:
                    Aclass.resid_contact_matrix_dict\
                            [(atom_name, atom_resname, atom_resid)]\
                            [(nAtom, nResname, nResid)] = 0
                    Aclass.resid_contact_matrix_dict\
                            [(nAtom, nResname, nResid)]\
                            [(atom_name, atom_resname, atom_resid)] = 0
                Aclass.resid_contact_matrix_dict\
                        [(atom_name, atom_resname, atom_resid)]\
                        [(nAtom, nResname, nResid)] += weight #1

            else:
                continue




def SorPopulation(Aclass, atom_name, atom_resname, 
        nearest_resname, RAD_str, RADshell_num, 
        RADshell_num_dist, waterTuple, N_H, 
        As, Ds, weight):


    Ds = tuple(Ds)
    As = tuple(As)
    DA_shell = Ds + As
    DA_shell = tuple(sorted(DA_shell))
    #print (DA_shell)


    for RR in \
            [[RADshell_num_dist, str(RADshell_num)], ['0', '0']]:

        RADshell_num_dist2 = RR[0]
        RADshell_num2 = RR[1]

        #### As and Ds as they are for pijs
        for ADs in [['A', As], ['D', Ds]]:
            if ADs[1] not in \
                    Aclass.RADshell_Sor_dict[nearest_resname]\
                    [(atom_resname, N_H, atom_name)]\
                    [(RADshell_num_dist2, str(RADshell_num2))]\
                    [RAD_str][ADs[0]]:

                Aclass.RADshell_Sor_dict[nearest_resname]\
                        [(atom_resname, N_H, atom_name)]\
                        [(RADshell_num_dist2, str(RADshell_num2))]\
                        [RAD_str]\
                        [ADs[0]][ADs[1]] = 0

            Aclass.RADshell_Sor_dict[nearest_resname]\
                    [(atom_resname, N_H, atom_name)]\
                    [(RADshell_num_dist2, str(RADshell_num2))]\
                    [RAD_str][ADs[0]][ADs[1]] += weight #1





def runningWeightedAverageFT(weight, weight_stored, count_stored, 
        forces_stored, torques_stored, T_stored, weighted_add, count_add, 
        forces_add, torques_add, T_add, 
        force, torque, KE):
    '''
    Get the running average of force and torque matrices
    '''


    try:
        if any(elem is None for elem in force) == False and \
                any(elem is None for elem in torque) == False:
            try:
                if torques_stored == 0 and forces_stored == 0:
                    weight_add = weight_stored + weight
                    count_add = count_stored + 1
                    forces_add = np.array(force)
                    torques_add = np.array(torque)
            except ValueError:
                weight_add = weight_stored + weight
                count_add = count_stored + 1
                #rolling_weighted_ave = (V_stored * w_stored + V * w) / 
                        #(w_stored + w)
                forces_add = ((forces_stored * weight_stored + 
                        force * weight) / float(weight_stored + weight))
                torques_add = ((torques_stored * weight_stored + 
                        torque * weight) / float(weight_stored + weight))

                if KE != None:
                    try:
                        DOF, T = 0, 0
                        for f in force[0]:
                            if f != 0:
                                DOF += 1
                            else:
                                continue

                        for t in torque[0]:
                            if t != 0:
                                DOF += 1
                            else:
                                continue

                        if DOF != 0:
                            T = KE / float(DOF) * 240.663461
                        if DOF == 0:
                            T = KE * 240.663461    
                        T_add = ((T + T_stored * count_stored) / 
                                float(count_add))
                    except TypeError: #force is a zero
                        pass

    except TypeError:
        pass

    return weight_add, count_add, forces_add, torques_add, T_add


def SvibPopulation(Aclass, level, atom_name, atom_resname, N_H, atom_num, 
        nearest_resname, RADshell_num_dist, waterTuple, 
        UAweightedForces, UAweightedTorques, PE, KE, MweightedForces, 
        MweightedTorques, Ndih, molecule_UA_F, molecule_UA_T, weight):



        #### DICTS FOR SVIB_ECALC

        ###switch level for solutes only to get WM Svib
        if level == 'res_atomLevel' and \
                atom_resname not in waterTuple:
            level == None
            atom_resname = atom_resname.split('_')[0]
        if level == 'residLevel_resname' and \
                atom_resname not in waterTuple:
            level == None
            nearest_resname = nearest_resname.split('_')[0] 
        atom_info = (atom_resname, N_H, atom_name)

        #print(nearest_resname, atom_resname)
        ###running weighted average of X

        if RADshell_num_dist not in Aclass.RADshell_dict\
                [nearest_resname][atom_info]:
            Aclass.RADshell_dict[nearest_resname]\
                    [atom_info]\
                    [RADshell_num_dist] = [0, 0, 0, 0, 0, 0, 0]

        if '0' not in Aclass.RADshell_dict\
                [nearest_resname][atom_info]:
            Aclass.RADshell_dict[nearest_resname]\
                    [atom_info]\
                    ['0'] = [0, 0, 0, 0, 0, 0, 0]

        for RADshell_num_dist2 in [RADshell_num_dist, '0']:
            forces_stored = Aclass.RADshell_dict[nearest_resname]\
                    [atom_info][RADshell_num_dist2][0]
            torques_stored = Aclass.RADshell_dict[nearest_resname]\
                    [atom_info][RADshell_num_dist2][1]
            PE_stored = Aclass.RADshell_dict[nearest_resname]\
                    [atom_info][RADshell_num_dist2][2]
            KE_stored = Aclass.RADshell_dict[nearest_resname]\
                    [atom_info][RADshell_num_dist2][3]
            T_stored = Aclass.RADshell_dict[nearest_resname]\
                    [atom_info][RADshell_num_dist2][4]
            count_stored = Aclass.RADshell_dict[nearest_resname]\
                    [atom_info][RADshell_num_dist2][5]
            weight_stored = Aclass.RADshell_dict[nearest_resname]\
                    [atom_info][RADshell_num_dist2][6]

            forces_add = forces_stored
            torques_add = torques_stored
            T_add = T_stored
            count_add = count_stored
            weight_add = weight_stored

            weight_add, count_add, forces_add, torques_add, T_add = \
                    runningWeightedAverageFT(weight, weight_stored, 
                    count_stored, forces_stored, 
                    torques_stored, T_stored, weight_add, count_add, 
                    forces_add, torques_add, T_add, 
                    UAweightedForces, UAweightedTorques, KE)

            PE_add = PE_stored
            KE_add = KE_stored

            if count_add != count_stored and PE != None:
                #PE_add = (PE + PE_stored * count_stored) / float(count_add)
                #KE_add = (KE + KE_stored * count_stored) / float(count_add)

                PE_add = ((PE_stored * weight_stored + 
                        PE * weight) / float(weight_stored + weight))
                KE_add = ((KE_stored * weight_stored + 
                        KE * weight) / float(weight_stored + weight))


            Aclass.RADshell_dict[nearest_resname]\
                    [atom_info][RADshell_num_dist2][0] = forces_add
            Aclass.RADshell_dict[nearest_resname]\
                    [atom_info][RADshell_num_dist2][1] = torques_add
            Aclass.RADshell_dict[nearest_resname]\
                    [atom_info][RADshell_num_dist2][2] = PE_add
            Aclass.RADshell_dict[nearest_resname]\
                    [atom_info][RADshell_num_dist2][3] = KE_add
            Aclass.RADshell_dict[nearest_resname]\
                    [atom_info][RADshell_num_dist2][4] = T_add
            Aclass.RADshell_dict[nearest_resname]\
                    [atom_info][RADshell_num_dist2][5] = count_add
            Aclass.RADshell_dict[nearest_resname]\
                    [atom_info][RADshell_num_dist2][6] = weight_add


        atom_info = (atom_info, Ndih)

        ##### PROX SHELL FT ASSIGNMENT

        if type(MweightedForces) != type(None):# and \
                #atom_resname in waterTuple: #water only for now

            if RADshell_num_dist not in Aclass.WM_FT_shell_dict\
                    [nearest_resname]\
                    [atom_info]:
                Aclass.WM_FT_shell_dict[nearest_resname]\
                        [atom_info]\
                        [RADshell_num_dist] = \
                        [0, 0, 0, 0, 0, 0, 0, 0, 0]

            forces_stored = Aclass.WM_FT_shell_dict[nearest_resname]\
                    [atom_info][RADshell_num_dist][0]
            torques_stored = Aclass.WM_FT_shell_dict[nearest_resname]\
                    [atom_info][RADshell_num_dist][1]
            PE_stored = Aclass.WM_FT_shell_dict[nearest_resname]\
                    [atom_info][RADshell_num_dist][2]
            KE_stored = Aclass.WM_FT_shell_dict[nearest_resname]\
                    [atom_info][RADshell_num_dist][3]
            T_stored = Aclass.WM_FT_shell_dict[nearest_resname]\
                    [atom_info][RADshell_num_dist][4]
            UAforces_stored = Aclass.WM_FT_shell_dict[nearest_resname]\
                    [atom_info][RADshell_num_dist][5]
            UAtorques_stored = Aclass.WM_FT_shell_dict[nearest_resname]\
                    [atom_info][RADshell_num_dist][6]
            count_stored = Aclass.WM_FT_shell_dict[nearest_resname]\
                    [atom_info][RADshell_num_dist][7]
            weight_stored = Aclass.WM_FT_shell_dict[nearest_resname]\
                    [atom_info][RADshell_num_dist][8]

            forces_add = forces_stored
            torques_add = torques_stored
            T_add = T_stored
            UAforces_add = UAforces_stored
            UAtorques_add = UAtorques_stored
            count_add = count_stored
            weight_add = weight_stored

            weight_add, count_add, forces_add, torques_add, T_add = \
                    runningWeightedAverageFT(weight, weight_stored, 
                    count_stored, forces_stored, 
                    torques_stored, T_stored, weight_add, count_add, 
                    forces_add, torques_add, T_add, 
                    MweightedForces, MweightedTorques, False)


            if len(molecule_UA_F) != 0 and len(molecule_UA_T) != 0:
                weight_add, count_ignore, UAforces_add, UAtorques_add, T_add = \
                        runningWeightedAverageFT(weight, weight_stored, 
                        count_stored, 
                        UAforces_stored, UAtorques_stored, 
                        T_stored, weight_add, count_add, 
                        UAforces_add, UAtorques_add, T_add, 
                        molecule_UA_F, molecule_UA_T, KE)

            PE_add = PE_stored
            KE_add = KE_stored

            if count_add != count_stored and PE != None:
                #PE_add = (PE + PE_stored * count_stored) / float(count_add)
                #KE_add = (KE + KE_stored * count_stored) / float(count_add)

                PE_add = ((PE_stored * weight_stored + 
                        PE * weight) / float(weight_stored + weight))
                KE_add = ((KE_stored * weight_stored + 
                        KE * weight) / float(weight_stored + weight))


            Aclass.WM_FT_shell_dict[nearest_resname]\
                    [atom_info][RADshell_num_dist][0] = forces_add
            Aclass.WM_FT_shell_dict[nearest_resname]\
                    [atom_info][RADshell_num_dist][1] = torques_add
            Aclass.WM_FT_shell_dict[nearest_resname]\
                    [atom_info][RADshell_num_dist][2] = PE_add
            Aclass.WM_FT_shell_dict[nearest_resname]\
                    [atom_info][RADshell_num_dist][3] = KE_add
            Aclass.WM_FT_shell_dict[nearest_resname]\
                    [atom_info][RADshell_num_dist][4] = T_add
            Aclass.WM_FT_shell_dict[nearest_resname]\
                    [atom_info][RADshell_num_dist][5] = UAforces_add
            Aclass.WM_FT_shell_dict[nearest_resname]\
                    [atom_info][RADshell_num_dist][6] = UAtorques_add
            Aclass.WM_FT_shell_dict[nearest_resname]\
                    [atom_info][RADshell_num_dist][7] = count_add
            Aclass.WM_FT_shell_dict[nearest_resname]\
                    [atom_info][RADshell_num_dist][8] = weight_add




def SconfPopulation(Aclass, nearest_resname, atom_name, atom_resname, N_H, 
        dihedral_phi_list, molecule_size, RADshell_num_dist, weight):
    '''
    '''

    def roundPartial(value, resolution):
        '''
        used for bin widths of 30 degrees
        '''
        return round(value / resolution) * resolution


    if nearest_resname != None and dihedral_phi_list != None:

        assigned = (atom_resname, len(dihedral_phi_list), 
                molecule_size, N_H, atom_name)

        counter = 0
        for dihedral_phi_type in dihedral_phi_list:
            #print (dihedral_phi_type)
            counter += 1
            #print ('c1:', counter)
            dihedral_atoms = (tuple(dihedral_phi_type[0]), counter)
                    #atomnum_list, dihedral index
            phi = dihedral_phi_type[1] #angle, degrees
            #dih_index = dihedral_phi_type[2][0] #0, 1, 2
            #dih_type = dihedral_phi_type[2][1] #trans, g-, g+

            ##adaptive method
            rounded_phi = roundPartial(phi, 30)
            #print (phi, rounded_phi)
            all_bins = list(range(-180, 181, 30))
            #print (all_bins)
            for b in all_bins:
                if b not in Aclass.adaptive_dih_dict\
                        [nearest_resname][assigned]\
                        [RADshell_num_dist][dihedral_atoms]:
                    Aclass.adaptive_dih_dict[nearest_resname]\
                            [assigned][RADshell_num_dist]\
                            [dihedral_atoms][b] = 0
            Aclass.adaptive_dih_dict[nearest_resname]\
                    [assigned][RADshell_num_dist]\
                    [dihedral_atoms][rounded_phi] += weight #1






#!/usr/bin/env python

import sys
import math
import numpy as np
import logging

from CodeEntropy.poseidon.extractData.generalFunctions import angle

from datetime import datetime


def UALevelRAD(all_data, dimensions):
    '''
    UA level RAD algorithm implementation.
    Iterate through the 20 closest UAs and find which ones
    are not blocked by a closer UA.
    '''

    for x in range(0, len(all_data)):
        i = all_data[x]
        if i.mass > 1.1:
            RAD = []
            RAD_dist = []
            if i.nearest_sorted_array != None:
                range_limit = len(i.nearest_sorted_array)
                if len(i.nearest_sorted_array) > 30:
                    range_limit = 30
                for y in range(0, range_limit):
                    j = all_data[i.nearest_sorted_array[y][0]]
                    rij = float(i.nearest_sorted_array[y][1])
                    blocked = False
                    for z in range(0, y): #only closer UAs can block
                        k = all_data[i.nearest_sorted_array[z][0]]
                        rik = float(i.nearest_sorted_array[z][1])
                        costheta_jik = float(angle(j.coords, i.coords, 
                                k.coords, dimensions))
                        if math.isnan(costheta_jik):
                           break 

                        LHS = (float(1)/rij) ** 2 
                        RHS = (((float(1)/rik) ** 2) * costheta_jik)
                        if LHS < RHS:
                            blocked = True
                            break

                    if blocked == False:
                        RAD.append(j)
                        RAD_dist.append((j, rij))

            i.RAD = RAD
            i.RAD_dist = RAD_dist
        else:
            continue




    '''
    Make sure each UA is in the other UAs RAD shell if not, remove
    check RAD shell contains resids other than the reference id
    (used for systems containing surfaces that are all the same resid)
    '''

    for x in range(0, len(all_data)):
        atom = all_data[x]
        if atom.mass > 1.1:
            if len(atom.RAD) != 0:
                RAD = [] #symmetric RAD shell here
                RAD_resids = [] #checking if neighbours are diff molecules
                #make sure atom is in neighbours RAD shell
                for neighbour in atom.RAD:
                    if atom in neighbour.RAD:
                        RAD.append(neighbour)
                        RAD_resids.append(neighbour.resid)
                    else:
                        continue

                if len(RAD) != 0:
                    ##when RAD shell contains one resid only
                    if all(n == RAD_resids[0] for n in RAD_resids) == True:
                        #make sure atom is not surrounded just by itself
                        if RAD_resids[0] == atom.resid:
                            atom.RAD = []
                        else:
                            atom.RAD = RAD
                    else:
                        atom.RAD = RAD
                if len(RAD) == 0:
                    atom.RAD = RAD
            else:
                continue
        else:
            continue



def distCutoffNc(all_data, dimensions, cutoff):
    '''
    Instead of RAD, use a fixed distance cutoff
    to define the first coordination shell (Nc)
    '''


    for x in range(0, len(all_data)):
        i = all_data[x]
        if i.mass > 1.1 and i.nearest_sorted_array != None:
            range_limit = len(i.nearest_sorted_array)
            if len(i.nearest_sorted_array) > 30:
                range_limit = 30
            RAD = []
            RAD_dist = []
            for y in range(0, range_limit):
                rij = float(i.nearest_sorted_array[y][1])
                if rij <= cutoff:
                    near = \
                        all_data[i.nearest_sorted_array\
                                [y][0]]
                    RAD.append(near)
                    RAD_dist.append((near, rij))
                else:
                    continue

            i.RAD = RAD
            i.RAD_dist = RAD_dist

        else:
            continue





def HBCalc(all_data, waterTuple, dimensions):
    '''
    HB TYPE CALC METHOD 2 (including all hydrogens (not just 
    for water) and finding most closest and most negative 
    atom to donate to - (QA*QD)/r^2 approach)

    ***addition = H's can only HB to UAs that are in the RAD
    shell of their bonded heavy atom.
    '''

    for x in range(0, len(all_data)):
        atom = all_data[x]
        if atom.mass < 1.1 and atom.mass >= 1 and atom.charge > 0.1:
            acceptor_charge = 99
            H = atom
            O = None
            for bonded in H.bonded_to_atom_num:
                b = all_data[bonded]
                if b.mass > 1.1:
                    O = b
                else:
                    continue

            ##Needed for RAD shell HB ranking, HBing inside RAD
            #RAD_atom_nums = [neighbour.atom_num for neighbour in O.RAD]
            RAD_atom_nums = []
            if O != None:
                for neighbour in O.RAD:
                    RAD_atom_nums.append(neighbour.atom_num)
                    for bonded in neighbour.bonded_to_atom_num:
                        b = all_data[bonded]
                        if b.mass < 1.1:
                            RAD_atom_nums.append(b.atom_num)
                        else:
                            continue

            if H.nearest_all_atom_array != None:
                range_limit = len(H.nearest_all_atom_array)
                if len(atom.nearest_all_atom_array) > 50:
                    range_limit = 50
                for atom_dist in H.nearest_all_atom_array[:range_limit]:
                    near = all_data[atom_dist[0]]
                    HX_dist = atom_dist[1]

                    if near.atom_num not in H.bonded_to_atom_num \
                            and near.atom_num != H.atom_num \
                            and near.charge != 0 \
                            and HX_dist != 0 and O != None \
                            and near.atom_num in RAD_atom_nums:

                        X = near
                        QD = H.charge
                        QA = X.charge

                        r2 = HX_dist ** 2
                        relative_charge = (float(QD) * float(QA)) / float(r2)
                        cosine_angle = angle(O.coords, H.coords, X.coords, 
                                    dimensions)
                        angle1 = np.arccos(cosine_angle)
                        OHX_angle = np.degrees(angle1)

                        if relative_charge < acceptor_charge \
                                and float(OHX_angle) > 90:
                            acceptor_charge = relative_charge
                            H.nearest_atom = X
                            H.dist = HX_dist

                        else:
                            continue
                    else:
                        continue
            else:
                continue

        else:
            continue


    broken_HBs = False
    #find hydrogens with neighbouring eneg atoms and append those 
    #hydrogens to the nearest_Hs of those eneg atoms
    for x in range(0, len(all_data)):
        H = all_data[x]
        if H.nearest_atom != None:
            nearest_eneg = H.nearest_atom.atom_num
            all_data[nearest_eneg].nearest_Hs.append(H)

        if broken_HBs:
            ####Dealing with broken HBs - ND, bifurcated and cyclic
            ####Hard-coded for water resnames only for now
            atom = all_data[x]
            HHX = False
            for b in atom.bonded_to_atom_num:
                bonded = all_data[b]
                if atom.mass < 1.1 and bonded.mass > 1.1:
                    if bonded.bondedUA_H != None:
                        if bonded.bondedUA_H[0] == 0 and \
                                bonded.bondedUA_H[1] == 2:
                            HHX = True

            if atom.mass < 1.1 and atom.nearest_all_atom_array != None and HHX:
                for atom_dist in atom.nearest_all_atom_array[:30]:
                    atom2 = all_data[atom_dist[0]]
                    bonded_overlap = bool(set(atom.bonded_to_atom_num) & 
                            set(atom2.bonded_to_atom_num)) 
                            #check if both bonded to same atom
                    if atom2.atom_num != atom.atom_num and \
                            atom2.resid == atom.resid and \
                            atom.nearest_atom != None and \
                            atom2.nearest_atom != None and \
                            bonded_overlap == True:
                        if atom2.nearest_atom.atom_num == \
                                atom.nearest_atom.atom_num \
                                and atom2.dist < atom.dist:

                            atom.broken = ['Bifurcated', atom.atom_num,
                                    atom.atom_name,
                                    atom.nearest_atom.atom_num, 
                                    atom.nearest_atom.atom_name] #####

                            '''
                            print ('Bifurcated', atom.atom_num,
                                    atom.atom_name,
                                    atom.nearest_atom.atom_num, 
                                    atom.nearest_atom.atom_name)
                            '''

                            for b in atom.bonded_to_atom_num:
                                bonded = all_data[b]
                                bonded.broken = ['Bifurcated', atom.atom_num,
                                    atom.atom_name,
                                    atom.nearest_atom.atom_num, 
                                    atom.nearest_atom.atom_name] #####

                            #remove this broken donor from acceptor
                            new_Hlist = []
                            for H in atom.nearest_atom.nearest_Hs:
                                if H != atom:
                                    new_Hlist.append(H)
                                else:
                                    continue

                            atom.nearest_atom.nearest_Hs = new_Hlist


                            for b2 in atom.nearest_atom.bonded_to_atom_num:
                                bonded2 = all_data[b2]
                                if bonded2.mass < 1.1:
                                    if bonded2.nearest_atom != None and \
                                            bonded2.nearest_atom.resid == \
                                            atom2.resid and \
                                            bonded2.dist < atom.dist:

                                        atom.broken = ['ND', atom.atom_num, 
                                                atom.atom_name, 
                                                atom.nearest_atom.atom_num, 
                                                atom.nearest_atom.atom_name] 
                                                #####

                                        '''
                                        print ('ND', atom.atom_num, 
                                                atom.atom_name, 
                                                atom.nearest_atom.atom_num, 
                                                atom.nearest_atom.atom_name)
                                        '''

                                        for b3 in atom.bonded_to_atom_num:
                                            bonded3 = all_data[b3]
                                            bonded3.broken = ['ND', 
                                                atom.atom_num, 
                                                atom.atom_name, 
                                                atom.nearest_atom.atom_num, 
                                                atom.nearest_atom.atom_name] 
                                                #####

                                        #remove this broken donor from acceptor
                                        new_Hlist = []
                                        for H in atom.nearest_atom.nearest_Hs:
                                            if H != atom:
                                                new_Hlist.append(H)
                                            else:
                                                continue

                                        atom.nearest_atom.nearest_Hs = new_Hlist
                                        atom.nearest_atom = None

                                    else:
                                        continue
                                else:
                                    continue

                            atom.nearest_atom = None


                    if atom2.atom_num != atom.atom_num and \
                            atom2.resid == atom.resid and \
                            bonded_overlap == True and \
                            len(atom2.nearest_Hs) != 0:
                        for H in atom2.nearest_Hs:
                            if atom.nearest_atom != None and \
                                    H.resid == atom.nearest_atom.resid and \
                                    H.dist < atom.dist:

                                bonded_overlap2 = bool(set(
                                    atom.nearest_atom.bonded_to_atom_num) \
                                    & set(H.bonded_to_atom_num)) 
                                    #check if both bonded to same atom

                                if bonded_overlap2 == True:

                                    atom.broken = ['Cyclic', atom.atom_num, 
                                            atom.atom_name, 
                                            atom.nearest_atom.atom_num, 
                                            atom.nearest_atom.atom_name] #####

                                    '''
                                    print ('Cyclic', atom.atom_num, 
                                            atom.atom_name, 
                                            atom.nearest_atom.atom_num, 
                                            atom.nearest_atom.atom_name)
                                    '''

                                    for b4 in atom.bonded_to_atom_num:
                                        bonded4 = all_data[b4]
                                        bonded4.broken = ['Cyclic', 
                                            atom.atom_num, 
                                            atom.atom_name, 
                                            atom.nearest_atom.atom_num, 
                                            atom.nearest_atom.atom_name] #####

                                    #remove this broken donor from acceptor
                                    new_Hlist = []
                                    for H in atom.nearest_atom.nearest_Hs:
                                        if H != atom:
                                            new_Hlist.append(H)
                                        else:
                                            continue

                                    atom.nearest_atom.nearest_Hs = new_Hlist
                                    atom.nearest_atom = None

                            else:
                                continue






#!/usr/bin/env python

import logging
import sys
import math
import numpy as np
from numpy import linalg as LA
from collections import Counter
from collections import defaultdict
import operator
from datetime import datetime

from CodeEntropy.poseidon.extractData.generalFunctions import *
from CodeEntropy.poseidon.extractData.mainClass import *

import MDAnalysis

nested_dict = lambda: defaultdict(nested_dict) 
        #create nested dict in one go

def populateTopology(container, all_data, waterTuple):
    '''
    After reading in topologies using MDAnalysis,
    relavant information from topology files are saved into a global
    class object. Properties are read in differently depeding on what
    file types were read in.
    '''

    all_resids = []
    mol = None

    for tp in container.atoms:
        bonded_atom_nums = []
        for b in tp.bonds:
            for x in b:
                if x.index != tp.index:
                    bonded_atom_nums.append(int(x.index))
                else:
                    continue

        try:
            dihedral_list = []
            if tp.mass > 1.1: 
                for dih in tp.dihedrals:
                    diha_list = []
                    for di in dih:
                        if di.mass > 1.1:
                            diha_list.append(int(di.index))
                        else:
                            break

                    if len(diha_list) == 4:
                        dihedral_list.append(diha_list)
                    else:
                        continue
        except AttributeError:
            dihedral_list = []


        inf = atom_info(int(tp.index), tp.name, tp.mass, 
                tp.charge, int(tp.resid), 
                tp.resname, bonded_atom_nums, dihedral_list)

        all_data.append(inf)

    ### Get and populate UA and molecule level information
    molecule_dict = nested_dict()
    molecule_resids_dict = nested_dict()
    for x in range(0, len(all_data)):
        atom = all_data[x]
        #print(f"x = {x}")
        if atom.mass > 1.1:
            heavy_bonded = []
            H_bonded = []
            for bonded_atom in atom.bonded_to_atom_num:
                #print(f"bonded_atom = {bonded_atom}")
                bonded = all_data[bonded_atom]
                if bonded.mass > 1.1:
                    heavy_bonded.append(bonded)
                elif bonded.mass < 1.1:
                    H_bonded.append(bonded)
                else:
                    continue

            bonded_atoms_list = [atom] + heavy_bonded + H_bonded
            atom.bondedUA_H = [len(heavy_bonded), len(H_bonded)]
           
            if atom.resid not in molecule_dict:
                molecule_dict[atom.resid] = []
                molecule_resids_dict[atom.resid] = []
            if atom.atom_num not in molecule_dict[atom.resid]:
                molecule_dict[atom.resid].append(atom.atom_num)
            for bonded2 in heavy_bonded:
                if bonded2.resid not in molecule_resids_dict[atom.resid]:
                    molecule_resids_dict[atom.resid].append(bonded2.resid)
                else:
                    continue
            

    for x in range(0, len(all_data)):
        atom = all_data[x]
        if atom.mass > 1.1:
            atom.molecule_atomNums = sorted(molecule_dict[atom.resid])
            atom.bonded_to_resids = sorted(molecule_resids_dict[atom.resid])
        else:
            continue





def getCoordsForces(container, all_data, dimensions, 
        frame, startTime, verbosePrint):
    '''
    Read in coordinate and force trajectories and populate mainClass.
    '''


    t = container.trajectory[frame]
    dimensions = np.array(t.dimensions[0:3])
    print(t)

    for x in range(0, len(t)):
        crds = np.array(t[x])
        all_data[x].coords = crds
        frcs = np.array(t.forces[x])
        all_data[x].forces = frcs
 
    verbosePrint('COORDS-FORCES')
    verbosePrint(datetime.now() - startTime)
    sys.stdout.flush() 

    return all_data, dimensions



# #Energy is not needed
# def populateEnergy(container, all_data, dimensions, frame, startTime, 
#         verbosePrint):
#     '''
#     read in energies from lammps input file.
#     '''

#     for e in container.trajectory:
#         if e.frame == frame:
#             dimensions = np.array(e.dimensions[0:3])
#             for x in range(0, len(e)):
#                 energy = np.array(e.velocities[x])
#                 all_data[x].PKenergy = [energy[0], energy[1]]
#             break
#         else:
#             continue


#     verbosePrint('ENERGIES')
#     verbosePrint(datetime.now() - startTime)
#     sys.stdout.flush() 





# def UAEnergyGroup(all_data):
#     '''
#     For energy on each atom, group together for each UA
#     and sum
#     '''

#     for x in range(0, len(all_data)):
#         atom = all_data[x]
#         if atom.mass > 1.1:
#             heavy_bonded = []
#             H_bonded = []
#             for bonded_atom in atom.bonded_to_atom_num:
#                 bonded = all_data[bonded_atom]
#                 if bonded.mass > 1.1:
#                     heavy_bonded.append(bonded)
#                 elif bonded.mass < 1.1:
#                     H_bonded.append(bonded)
#                 else:
#                     continue

#             bonded_atoms_list = [atom] + heavy_bonded + H_bonded
#             atom.bondedUA_H = [len(heavy_bonded), len(H_bonded)]
#             UA_atoms_list = [atom] + H_bonded

#             UA_PE_list = []
#             UA_KE_list = []
#             for A in UA_atoms_list:
#                 if A.PKenergy != None:
#                     UA_PE_list.append(A.PKenergy[0])
#                     UA_KE_list.append(A.PKenergy[1])
#                 else:
#                     continue

#             if len(UA_PE_list) != 0:
#                 UA_PE = round(sum(UA_PE_list), 3)
#                 UA_KE = round(sum(UA_KE_list), 3)
#                 atom.UA_PKenergy = [UA_PE, UA_KE]






def getDistArray(atom, all_data, traj, max_cutoff,
        dimensions, neighbour_coords, startTime, verbosePrint):
    '''
    Find the NN list of an atom
    Important to use coords directly from MDAnalysis to run NN calc
    '''

    atom_coords = traj[atom.atom_num]

    #added a small min cutoff to stop zero distance
    array1, array2 = \
            MDAnalysis.lib.distances.capped_distance(atom_coords, 
                    neighbour_coords, max_cutoff=max_cutoff, 
                    min_cutoff=None, box=traj.dimensions, 
                    method=None, return_distances=True)


    try:
        array1, array2 = zip(*sorted(zip(array2, array1), 
            key=lambda x: x[0]))

    except ValueError:
        logging.error('Bad Arrays for Coordinate/ Atom Number '\
                'Nearest Neighbour Assignments')

    atomNumList = []
    allAtomList = []
    for atoms, dist in zip(array2, array1):
        near = atoms[1]
        atom_num = all_data[near].atom_num
        atom_resid = all_data[near].resid
        atom_mass = all_data[near].mass
        allAtomList.append((near, dist))

        #atom_resid != all_data[x].resid removed for quartz
        #surface that is all the same resid.
        if atom_num != atom.atom_num \
                and atom_num not in \
                    atom.bonded_to_atom_num \
                and float(atom_mass) > 1.1:
            atomNumList.append((near, dist))
        else:
            continue

    atom.nearest_sorted_array = atomNumList
    atom.nearest_all_atom_array = allAtomList






#!/usr/bin/env python

import math
import numpy as np
import sys

from CodeEntropy.poseidon.extractData.generalFunctions import calcAngleWithNearestNonlike


def moleculeObjectPopulation(all_data, allMoleculeList, frame, dimensions):
    '''
    For all molecules in the system, populate a list with required
    info for further analysis across any trajectory frame.
    '''


    for x in range(0, len(all_data)):
        atom = all_data[x]
        if atom.mass > 1.1:
            nearestInfo = []
            if atom.nearestAnyNonlike != None:
                nearestInfo = [
                    atom.nearestAnyNonlike[0].atom_name,
                    atom.nearestAnyNonlike[0].resname,
                    atom.nearestAnyNonlike[0].resid,
                    round(atom.nearestAnyNonlike[1], 3) #dist
                    ]

            allMoleculeList.append([
                frame, #0
                atom.atom_name, #1
                atom.atom_num, #2
                atom.resname, #3
                atom.resid, #4
                atom.bondedUA_H, #5
                atom.molecule_atomNums, #6
                nearestInfo, #7
                atom.hydrationShellRAD_solute, #8
                atom.RAD_shell_ranked, #9
                atom.MweightedForces, #10
                atom.MweightedTorques, #11
                atom.UA_PKenergy, #12
                atom.UAweightedForces, #13
                atom.UAweightedTorques, #14
                atom.dihedral_phi_type, #15
                atom.molecule_UA_Fs, #16
                atom.molecule_UA_Ts, #17
                atom.RAD_nAtoms, #18
                ])


        else:
            continue


    return allMoleculeList





def pdbGenerate(all_data, fileName, dimensions):
    '''
    Output a pdb file to visualise different waters in system.
    Might have to also make an equivalent .csv file so that
    we have an easier time analysing the files.
    But append to one global file rather than each frame, 
    so need a column of frame number for analysis later.
    pdb file pure for producing images with vmd, 
    but could create these later anyway? 
    Then we can choose our beta column?
    '''

    data = open(fileName+'_solProx_dist.pdb', "w") 
        #solute resid, int(dist) from solute resid

    data3 = open(fileName+'_RADHBlengths.pdb', "w") 
        #Nc, N_acceptors/N_donors

    data6 = open(fileName+'_solProx_RAD.pdb', "w") 
        #solute resid, int(dist) from solute resid

    file_list = [data, data3, data6]
    
    for x in range(0, len(all_data)):
        atom = all_data[x]


        try:
            a = int(atom.nearestAnyNonlike[0].resid)
        except TypeError:
            a = 0
        try:
            b = int(atom.nearestAnyNonlike[1])
        except TypeError:
            b = 0
        try:
            e = len(atom.RAD)
        except TypeError:
            e = 0
        try:
            As = len(atom.RAD_shell_ranked[1])
            Ds = len(atom.RAD_shell_ranked[2])
            DAhh_type = '%s%s' % (Ds, As)
            f = str(DAhh_type)
        except TypeError:
            f = '00'
        try:
            i = int(atom.hydrationShellRAD_solute)
        except (TypeError, ValueError):
            i = 0

        data.write("\n".join(["{:6s}{:5d} {:^4s}{:1s}{:3s}" \
                " {:1s}{:4d}{:1s}   {:8.3f}{:8.3f}{:8.3f}" \
                "{:6.0f}{:6.0f}          {:>2s}{:2s}".format(
                    "ATOM", int(str(atom.atom_num)[0:5]), 
                    atom.atom_name, " ", atom.resname[0:3], 
                    "A", int(str(atom.resid)[0:4]), " ", 
                    float(atom.coords[0]), 
                    float(atom.coords[1]), 
                    float(atom.coords[2]), 
                    a, b, " ", " ")]) + "\n")

        data3.write("\n".join(["{:6s}{:5d} {:^4s}{:1s}{:3s}" \
                " {:1s}{:4d}{:1s}   {:8.3f}{:8.3f}{:8.3f}" \
                "{:6.0f}    {:2s}          {:>2s}{:2s}".format(
                    "ATOM", int(str(atom.atom_num)[0:5]), 
                    atom.atom_name, " ", atom.resname[0:3], 
                    "A", int(str(atom.resid)[0:4]), " ", 
                    float(atom.coords[0]), 
                    float(atom.coords[1]), 
                    float(atom.coords[2]), 
                    e, f, " ", " ")]) + "\n")
        
        data6.write("\n".join(["{:6s}{:5d} {:^4s}{:1s}{:3s}" \
                " {:1s}{:4d}{:1s}   {:8.3f}{:8.3f}{:8.3f}" \
                "{:6.0f}{:6.0f}          {:>2s}{:2s}".format(
                    "ATOM", int(str(atom.atom_num)[0:5]), 
                    atom.atom_name, " ", atom.resname[0:3], 
                    "A", int(str(atom.resid)[0:4]), " ", 
                    float(atom.coords[0]), 
                    float(atom.coords[1]), 
                    float(atom.coords[2]), 
                    a, i, " ", " ")]) + "\n")

    for d in file_list:
        d.close()





import xdrlib
import numpy as nmp

def write_FTC_matrix_bin( arg_beadCollection ):
	""" This function writes the FTC matrix (FF + FT + TF + TT) for a given instance of BeadCollection in binary format.
	The sequence of the data is:
	N (# Beads), M x M (length of FTC matrix), MFF x MFF (length of FF matrix), MFT_row x MFT_col (length of FT matrix),
	MTF_row x MTF_col (length of TF matrix), MTT x MTT (length of TT matrix),
	FF matrix as 1D array, FT matrix as 1D array, TF matrix as 1D array, TT matrix as 1D array. Dimensions are unsigned int
	and arrays are in floats """

	fBin = open('FTC_matrix.bin', 'wb')

	xpa = xdrlib.Packer()

	# number of beads
	xpa.pack_uint( len(arg_beadCollection.listOfBeads) )

	# dim of FTC matrix
	pack_matrix_shape( arg_beadCollection.forceTorqueCovMatrix, xpa )
	
	# dim of FF, FT, TF, TT
	for iKey in arg_beadCollection.subMatrixDict.keys():
		iMat = arg_beadCollection.generate_quadrant( iKey, True)
		pack_matrix_shape(iMat, xpa)

	# flatten the submatrices and store them in 1D arrays
	for iKey in arg_beadCollection.subMatrixDict.keys():
		iMat = arg_beadCollection.generate_quadrant(iKey, True)
		pack_matrix_as_array(iMat, xpa)

	# write he buffer into fBin
	fBin.write(xpa.get_buffer())

	fBin.close()

	return
#END

def pack_matrix_shape( arg_matrix, arg_packer ):
	for iSize in nmp.shape(arg_matrix):
		# print("packing", iSize)
		arg_packer.pack_uint(iSize)

	return
#END

def pack_matrix_as_array ( arg_matrix, arg_packer ):
	for iVal in arg_matrix.flatten():
		arg_packer.pack_float(iVal)

	return
#END





import numpy as nmp
import sys
from collections import deque
from CodeEntropy.FunctionCollection import Utils

class LinkedListNode(object):
	def __init__(self, arg_value):
		self.value = arg_value
		self.nextNode = None

	def __eq__(self, arg_otherNode):
		return self.value == arg_otherNode.value

class DoublyLinkedListNode(LinkedListNode):
	def __init__(self, arg_value):
		super().__init__(arg_value)
		self.previousNode = None

	def __iter__(self):
		return self

class CircularLinkedList(object):
	""" 
	A linked list where each node is connected to its next node 
	and the last node is linked with the first node,
	thus completing a circle"""

	def __init__(self):
		self.currentPosition = None
		self.startPosition = None
		self.count = 0

	def append(self, arg_node):
		if self.count == 0:
			self.startPosition = arg_node
			self.currentPosition = arg_node
			arg_node.nextNode = self.currentPosition
			self.incr_count()

		else:
			self.currentPosition.nextNode = arg_node
			self.currentPosition = arg_node
			arg_node.nextNode = self.startPosition
			self.incr_count()

		return

	def incr_count(self):
		self.count += 1
		return

	def reset_position(self):
		self.currentPosition = self.startPosition
		return

	def __len__(self):
		return self.count

	def get_current_position(self):
		return self.currentPosition

	def move_right_of_current_position(self):
		self.currentPosition = self.currentPosition.nextNode
		return
#END

class DoublyLinkedList(object):
	""" Each node is connected to its next and previous node."""
	def __init__(self):
		self.startPosition = None
		self.endPosition = None
		self.currentPosition = None
		self.count = 0


	def append(self, arg_doubleNode):
		if self.count == 0:
			# means it points to nothing
			# initialize it now
			self.startPosition = arg_doubleNode
			self.endPosition = arg_doubleNode
			self.currentPosition = arg_doubleNode
			self.incr_count()

		else:
			# add it to the right of the currently last position
			self.endPosition.nextNode = arg_doubleNode
			self.endPosition = arg_doubleNode
			self.currentPosition = self.endPosition
			self.incr_count()

		return

	def incr_count(self):
		self.count += 1
		return

	def decr_count(self):
		self.count -= 1
		assert(self.count >= 0)
		return

	def __len__(self):
		return self.count

	def get_current_position(self):
		return self.currentPosition

	def get_node_at(self, arg_index):
		#0-indexed
		try:
			assert(arg_index < self.count and arg_index >= 0)
		except:
			raise IndexError('{} exceeds the last index of the list ({})'.format(arg_index, self.count - 1))
		
		# backup the current postion
		backupCurrentPosition = self.get_current_position()

		idx = 0
		self.move_to_start()
		idxNode = self.get_current_position()
		while idx != arg_index:
			idxNode = idxNode.nextNode
			idx += 1

		self.currentPosition = backupCurrentPosition
		return idxNode


	def move_to_start(self):
		self.currentPosition = self.startPosition
		return 1

	def move_to_end(self):
		self.currentPosition = self.endPosition
		return 1

	def move_right_of_current_position(self):
		self.currentPosition = self.currentPosition.nextNode
		return 

	def move_left_of_current_position(self):
		self.currentPosition = self.currentPosition.previousNode
		return 

	def remove_node_and_move_right(self):
		""" Remove the node pointed by the currentPosition and then move to its nextNode """
		leftNode = self.currentPosition.previousNode
		rightNode = self.currentPosition.nextNode

		if leftNode != None:
			leftNode.nextNode = rightNode
		if rightNode != None:
			rightNode.previousNode = leftNode

		self.currentPosition = rightNode
		self.decr_count()
		return 1

	def remove_node_and_move_left(self):
		""" Remove the node pointed by the currentPosition and then move to its previousNode """
		leftNode = self.currentPosition.previousNode
		rightNode = self.currentPosition.nextNode

		if leftNode != None:
			leftNode.nextNode = rightNode
		if rightNode != None:
			rightNode.previousNode = leftNode

		self.currentPosition = previousNode
		self.decr_count()
		return 1

	def clear_all(self):
		""" Basically, point all its memebr pointers to None and reset its count to 0 """
		self.startPosition = None
		self.endPosition = None
		self.currentPosition = None
		self.count = 0
		return


	def iterator(self):
		assert(self.count != 0)

		self.move_to_start()
		while self.currentPosition:
			yield self.currentPosition
			self.move_right_of_current_position()
#END

class CircularDoublyLinkedList(DoublyLinkedList):
	""" 
	Each node is connected to its next and previous node. At
	the end of the complete circle, the next node to the last 
	node is the first node. This is a derived class of the 
	`DoublyLinkedList` class.
	"""

	def __init__(self):
		super().__init__()


	def append(self, arg_doubleNode):
		if self.count == 0:
			# means it points to nothing
			# initialize it now
			self.startPosition = arg_doubleNode
			self.endPosition = arg_doubleNode
			self.currentPosition = arg_doubleNode
			self.incr_count()

		else:
			# add it to the right of the currently last position
			self.endPosition.nextNode = arg_doubleNode
			self.startPosition.previousNode = arg_doubleNode

			arg_doubleNode.previousNode = self.endPosition
			arg_doubleNode.nextNode = self.startPosition

			self.endPosition = arg_doubleNode
			self.currentPosition = self.endPosition

			self.incr_count()

		return

	def incr_count(self):
		self.count += 1
		return

	def decr_count(self):
		self.count -= 1
		assert(self.count >= 0)
		return

	def __len__(self):
		return self.count

	def get_current_position(self):
		return self.currentPosition

	def get_node_at(self, arg_index):
		#0-indexed
		try:
			assert(arg_index < self.count and arg_index >= 0)
		except:
			raise IndexError('{} exceeds the last index of the list ({})'.format(arg_index, self.count - 1))
		
		# backup the current postion
		backupCurrentPosition = self.get_current_position()

		idx = 0
		self.move_to_start()
		idxNode = self.get_current_position()
		while idx != arg_index:
			idxNode = idxNode.nextNode
			idx += 1

		self.currentPosition = backupCurrentPosition
		return idxNode


	def move_to_start(self):
		self.currentPosition = self.startPosition
		return 1

	def move_to_end(self):
		self.currentPosition = self.endPosition
		return 1

	def move_right_of_current_position(self):
		self.currentPosition = self.currentPosition.nextNode
		return 

	def move_left_of_current_position(self):
		self.currentPosition = self.currentPosition.previousNode
		return 

	def remove_node_and_move_right(self):
		""" Remove the node pointed by the currentPosition and then move to its nextNode """
		leftNode = self.currentPosition.previousNode
		rightNode = self.currentPosition.nextNode

		if leftNode != None:
			leftNode.nextNode = rightNode
		if rightNode != None:
			rightNode.previousNode = leftNode

		self.currentPosition = rightNode
		self.decr_count()
		return 1

	def remove_node_and_move_left(self):
		""" Remove the node pointed by the currentPosition and then move to its previousNode """
		leftNode = self.currentPosition.previousNode
		rightNode = self.currentPosition.nextNode

		if leftNode != None:
			leftNode.nextNode = rightNode
		if rightNode != None:
			rightNode.previousNode = leftNode

		self.currentPosition = previousNode
		self.decr_count()
		return 1

	def clear_all(self):
		""" Basically, point all its memebr pointers to None and reset its count to 0 """
		self.startPosition = None
		self.endPosition = None
		self.currentPosition = None
		self.count = 0
		return


	def iterator_cw(self):
		"""
		Iterate through next nodes (clockwise) from the start to the end
		of the circular doubly linked list.
		"""
		# only iterate if the list is non-empty
		assert(self.count != 0)

		# create a backup of the current poisition to restore it later
		backupCurrentPosition = self.currentPosition

		# begin iteration
		self.move_to_start()
		while self.currentPosition:
			yield self.currentPosition
			if self.currentPosition == self.endPosition:
				break
			else:
				self.move_right_of_current_position()

		# restore the original current position after iteration is completed
		self.currentPosition = backupCurrentPosition

	def iterator_ccw(self):
		"""
		Iterate through previous nodes (counter-clockwise) from the start to the end
		of the circular doubly linked list.
		"""
		# only iterate if the list is non-empty
		assert(self.count != 0)

		# create a backup of the current poisition to restore it later
		backupCurrentPosition = self.currentPosition

		# begin iteration
		self.move_to_end()
		while self.currentPosition:
			yield self.currentPosition
			if self.currentPosition == self.startPosition:
				break
			else:
				self.move_left_of_current_position()

		# restore the original current position after iteration is completed
		self.currentPosition = backupCurrentPosition

#END

class TreeNode(object):
	""" A generic data structure that depicts a node for a binary tree. 
	Has a pointer to one parent and 2 children """

	def __init__(self, arg_leftChildNode, arg_rightChildNode, arg_value):
		self.leftChildNode = arg_leftChildNode
		self.rightChildNode = arg_rightChildNode
		self.value = arg_value

	def __lt__(self, arg_otherNode):
		assert(self.value != None and arg_otherNode.value != None)
		return self.value < arg_otherNode.value

	def __gt__(self, arg_otherNode):
		assert(self.value != None and arg_otherNode.value != None)
		return self.value > arg_otherNode.value


#END

class BinaryTree(object):
	""" A rooted bindary tree which will be defined by a root tree node. It will be populated by links 
	to the root nodes and their subsequent links"""
	def __init__(self):
		self.rootNode = None
		self.nodeCount = 0

	def __len__(self):
		return self.nodeCount
	#END


	def add_node(self, arg_node):
		if self.rootNode == None:
			# arg_node is the root, add it likewise
			self.rootNode = arg_node
			self.incr_nodeCount()
			return

		else:
			self.add_node_to_rooted_tree(arg_parentNode = self.rootNode, arg_node = arg_node)
			return
	#END

	def add_node_to_rooted_tree(self, arg_parentNode, arg_node):
		if arg_node < arg_parentNode:
			if arg_parentNode.leftChildNode == None:
				arg_parentNode.leftChildNode = arg_node
				arg_node.parentNode = arg_parentNode
				self.incr_nodeCount()
				return
			else:
				self.add_node_to_rooted_tree(arg_parentNode = arg_parentNode.leftChildNode, arg_node = arg_node)

		elif arg_node > arg_parentNode:
			if arg_parentNode.rightChildNode == None:
				arg_parentNode.rightChildNode =arg_node
				arg_node.parentNode = arg_parentNode
				self.incr_nodeCount()
				return
			else:
				self.add_node_to_rooted_tree(arg_parentNode = arg_parentNode.rightChildNode, arg_node = arg_node)

	def list_in_order(self):
		outList = []
		try:
			assert(self.rootNode != None)
		except:
			# print('Tree is non-existent. It will return an empty list')
			return outList

		self.generate_IN_order_list_from_node(arg_node = self.rootNode, arg_outList = outList)
		return outList

	#END	



	def generate_IN_order_list_from_node(self, arg_node, arg_outList):
		if arg_node.leftChildNode == None:
			pass
		else:
			self.generate_IN_order_list_from_node(arg_node = arg_node.leftChildNode, arg_outList = arg_outList)

		arg_outList.append(arg_node.value) 

		if arg_node.rightChildNode == None:
			pass
		else:
			self.generate_IN_order_list_from_node(arg_node = arg_node.rightChildNode, arg_outList = arg_outList) 

		return arg_outList
	#END


	
	def incr_nodeCount(self):
		# Utils.printflush('Increasing node count from {}'.format(self.nodeCount), end = '')
		self.nodeCount += 1
		# Utils.printflush(' to {}'.format(self.nodeCount))
		return

	#END

	def decr_nodeCount(self):
		# Utils.printflush('Decreasing node count from {}'.format(self.nodeCount), end = '')
		self.nodeCount -= 1
		# Utils.printflush(' to {}'.format(self.nodeCount))
		return
	#END

#END

class AxesSystem(object):
	"""
	A class that contains the orthogonal axes and 
	the origin of a N-dimensional space.
	"""

	def __init__(self, arg_ndim):
		self.basisVectors = nmp.zeros((arg_ndim, arg_ndim))
		self.origin = nmp.zeros(arg_ndim)

	def __str__(self):
		retStr = "Basis vectors:\n"
		for i in range(self.get_dimension()):
			retStr = "{}|".format(retStr)
			for j in range(self.get_dimension()):
				retStr = "{}{:>8.3f}".format(retStr, self.basisVectors[i][j])
			retStr = "{}|\n".format(retStr)

		retStr = "{}\n\nOrigin:\n"
		for k in range(self.get_dimension()):
			retStr = "{}{:>8.3f}".format(retStr, self.origin[k])
		return retStr

	def get_basis_vectors(self):
		return self.basisVectors

	def get_dimension(self):
		""" Return the dimensionality of the axes system."""
		return len(self.origin)

	def transpose(self):
		""" Return a matrix where the columns are system's basis vectors.""" 
		return nmp.tramspose(self.basisVectors)










#!/usr/bin/env python

import os
import sys
import psutil

import math
import numpy as np


from collections import Counter

from CodeEntropy.poseidon.analysis.classes import *
from CodeEntropy.poseidon.analysis.populateEE import SorPopulation, SvibPopulation, \
        SconfPopulation, contactPopulation, \
        contactPopulationUA


def classPopulation(atomList, entropyEnergy, 
        level_list, count, atom_count, 
        num_atoms, waterTuple, 
        temperature,  
        solvent, 
        EEclass, EEclass_residLevel_resname, EEclass_soluteContacts, 
        EEclass_atomLevel,
        weighting_info, verbosePrint):
    '''
    For each flag, populate dictionaries accordingly
    The objects read in are used for this.
    '''

    ##### Smix_methods
    step_list = None

    #'''
    if count == 1: 

        #initiate classes for each flag

        if entropyEnergy:
            EEclass = variables_dicts(True)
            EEclass_residLevel_resname = variables_dicts(True)
            EEclass_soluteContacts = variables_dicts(True)
            EEclass_atomLevel = variables_dicts(True)


        if weighting_info != None:
            if EEclass == None:
                EEclass = variables_dicts(True)
            EEclass.weighting_list = weighting_info[0]
            EEclass.weighting_chunks = weighting_info[1]


    #'''


    for i in atomList: #iterate through each atom
        ####### only need to change this if list order changes (and below)
        atom_count += 1
        globalFrame = math.ceil(atom_count / float(num_atoms)) #round UP
        frame = i[0] #not used
        cumulative_steps = ['All']
        if step_list != None:
            for n in range(1, len(step_list)):
                step = step_list[n]
                if int(globalFrame) <= step:
                    cumulative_step = str(step)
                    #print(globalFrame, cumulative_step)
                    cumulative_steps.append(cumulative_step)

        atom_name = i[1]
        atom_num = i[2]
        atom_resname = i[3]
        atom_resid = i[4]
        N_H = i[5][1]
        #molecule_atoms = i[6] #heavy atoms only
        molecule_size = len(i[6]) #heavy atoms only
        try:
            nearest_atomName = i[7][0]
            nearest_resname = i[7][1]
            nearest_resid = i[7][2]
            #dist = float(i[7][3])
            #dist = round(float(i[7][3]), 1)
        except IndexError:
            nearest_atomName = 'unassigned'
            nearest_resname = 'unassigned'
            nearest_resid = 0
            #dist = 0
        RADshell_num = i[8]
        if RADshell_num == None:
            RADshell_num = '0' #unassigned to a nearest
        RAD_list = sorted(i[9][0])
        As = i[9][1]
        Ds = i[9][2]
        ### Svib_Ecalcs
        MweightedForces = i[10] #array
        MweightedTorques = i[11] #array
        PE = None
        KE = None
        if i[12] != None:
            PE = i[12][0]
            KE = i[12][1]

        UAweightedForces = i[13]
        UAweightedTorques = i[14]
        #### FT POPULATION
        try:
            dihedral_phi_list = i[15]
            Ndih = len(i[15])
        except (IndexError, TypeError):
            dihedral_phi_list
            Ndih = None
        try:
            molecule_UA_F = i[16]
            molecule_UA_T = i[17]
        except IndexError:
            Ndih = None
            molecule_UA_F = []
            molecule_UA_T = []
        RAD_nAtoms = i[18] #closest atom in each RADshell and distance

        RAD_str = tuple(sorted(RAD_list))

        ### toggle between shells or distances as x variable
        RADshell_num_dist = str(RADshell_num)

        #### group all solute types together for easier analysis
        if atom_resname not in solvent and RADshell_num != '0':
            nearest_resname = 'Any'
            RADshell_num_dist = '1'


        #### find out weighting of each frame if simulation is biased
        weight = 1
        if weighting_info != None:
            if EEclass.weighting_list != None:
                weight = getWeight(EEclass, globalFrame)

        for level in level_list:

            nearest_resname_list = []
            atom_resnameLevel, nearest_resname_list, waterTuple \
                    = levelSelection(level, 
                    atom_resname, atom_resid, nearest_resname, 
                    nearest_resname_list, atom_name, 
                    nearest_atomName, nearest_resid, RAD_nAtoms, waterTuple)


            for nearest_resnameLevel in nearest_resname_list:

                if 'Any' in nearest_resnameLevel:
                    nearest_resnameLevel = 'Any'


                if level in [None, 'moleculeLevel']:

                    if entropyEnergy:

                        SorPopulation(EEclass, 
                                atom_name, atom_resnameLevel, 
                                nearest_resnameLevel, RAD_str, 
                                RADshell_num, RADshell_num_dist, 
                                waterTuple, N_H, 
                                As, Ds, weight)

                        SvibPopulation(EEclass, level, atom_name, 
                                atom_resnameLevel, 
                                N_H, atom_num, 
                                nearest_resnameLevel, 
                                RADshell_num_dist, waterTuple, 
                                UAweightedForces, UAweightedTorques, PE, KE, 
                                MweightedForces, 
                                MweightedTorques, Ndih, molecule_UA_F, 
                                molecule_UA_T, weight)

                        SconfPopulation(EEclass, nearest_resnameLevel, 
                                atom_name, atom_resnameLevel, 
                                N_H, dihedral_phi_list, molecule_size, 
                                RADshell_num_dist, weight)



                if level in ['residLevel_resname']:
                    if entropyEnergy:

                        contactPopulation(EEclass_residLevel_resname, 
                                RAD_nAtoms, 
                                atom_resnameLevel, atom_resid, waterTuple, 
                                weight)

                        SorPopulation(EEclass_residLevel_resname, 
                                atom_name, atom_resnameLevel, 
                                nearest_resnameLevel, RAD_str, 
                                RADshell_num, RADshell_num_dist, 
                                waterTuple, N_H, 
                                As, Ds, weight)

                        SvibPopulation(EEclass_residLevel_resname, 
                                level, atom_name, atom_resnameLevel, 
                                N_H, atom_num, 
                                nearest_resnameLevel, 
                                RADshell_num_dist, waterTuple, 
                                UAweightedForces, UAweightedTorques, PE, KE, 
                                MweightedForces, 
                                MweightedTorques, Ndih, molecule_UA_F, 
                                molecule_UA_T, weight)

                        SconfPopulation(EEclass_residLevel_resname, 
                                nearest_resnameLevel, 
                                atom_name, atom_resnameLevel, 
                                N_H, dihedral_phi_list, molecule_size, 
                                RADshell_num_dist, weight)


                if level in ['soluteContacts']:

                    if entropyEnergy:
                        SorPopulation(EEclass_soluteContacts, 
                                atom_name, atom_resnameLevel, 
                                nearest_resnameLevel, RAD_str, 
                                RADshell_num, RADshell_num_dist, 
                                waterTuple, N_H, 
                                As, Ds, weight)

                        SvibPopulation(EEclass_soluteContacts, 
                                level, atom_name, atom_resnameLevel, 
                                N_H, atom_num, 
                                nearest_resnameLevel, 
                                RADshell_num_dist, waterTuple, 
                                UAweightedForces, UAweightedTorques, PE, KE, 
                                MweightedForces, 
                                MweightedTorques, Ndih, molecule_UA_F, 
                                molecule_UA_T, weight)

                        SconfPopulation(EEclass_soluteContacts, 
                                nearest_resnameLevel, 
                                atom_name, atom_resnameLevel, 
                                N_H, dihedral_phi_list, molecule_size, 
                                RADshell_num_dist, weight)



                if level in ['atomLevel']:

                    if entropyEnergy:

                        contactPopulationUA(EEclass_atomLevel, 
                                RAD_nAtoms, atom_name,
                                atom_resnameLevel, atom_resid, waterTuple, 
                                weight)

                        SorPopulation(EEclass_atomLevel, 
                                atom_name, atom_resnameLevel, 
                                nearest_resnameLevel, RAD_str, 
                                RADshell_num, RADshell_num_dist, 
                                waterTuple, N_H, 
                                As, Ds, weight)

                        SvibPopulation(EEclass_atomLevel, 
                                level, atom_name, atom_resnameLevel, 
                                N_H, atom_num, 
                                nearest_resnameLevel, 
                                RADshell_num_dist, waterTuple, 
                                UAweightedForces, UAweightedTorques, PE, KE, 
                                MweightedForces, 
                                MweightedTorques, Ndih, molecule_UA_F, 
                                molecule_UA_T, weight)

                        SconfPopulation(EEclass_atomLevel, 
                                nearest_resnameLevel, 
                                atom_name, atom_resnameLevel, 
                                N_H, dihedral_phi_list, molecule_size, 
                                RADshell_num_dist, weight)

    sys.stdout.flush() 

    return EEclass, EEclass_residLevel_resname, \
            EEclass_soluteContacts, EEclass_atomLevel, \
            atom_count




def getWeight(EEclass, globalFrame):
    '''
    Find out what the weighting is for a given simualtion frame
    '''

    #chunk = int(EEclass.weighting_chunks)
    weighting_list = EEclass.weighting_list

    index = globalFrame-1
    try:
        weight = float(weighting_list[index])
        #print('weight: %s' % (weight))
    except (IndexError, ValueError):
        weight = 1

    #print(globalFrame, index, weight)

    return weight




def stripNumbers(atom_name, atom_resname):
    '''
    '''

    atom_name_letters = ''.join([i for i in atom_name 
            if not i.isdigit()])
    if atom_name != atom_resname:
        atom_name_letters = atom_name_letters[0]
    atom_resname = ('%s_%s' % (atom_resname, atom_name_letters))

    return atom_resname






def levelSelection(level, atom_resname, atom_resid, nearest_resname, 
        nearest_resname_list, 
        atom_name, nearest_atomName, nearest_resid, RAD_nAtoms, waterTuple):
    '''
    Depending on what level of analysis we want to do, dictionaries are
    populated according to how the nearest and assigned are defined.

    For example, we may want to analyse molecule types, the atoms 
    in each molecule, molecules by residue number or multiple solutes 
    in a RAD shell as a list of nearest molecules.
    '''

    ###keep as molecule
    if level == 'moleculeLevel' or level == None:
        atom_resname = atom_resname
        nearest_resname_list = [nearest_resname]


    ###use resname_atomName for both nearest and assigned
    if level == 'atomLevel': #res_atomLevel
        #'''
        water = False
        if atom_resname in waterTuple:
            water = True
        atom_resname = stripNumbers(atom_name, atom_resname)
        if atom_resname not in waterTuple and water:
            waterTuple.append(atom_resname)
        nearest_resname_list = [stripNumbers(nearest_atomName, 
            nearest_resname)]

        #'''

    '''
    ###Inclusion of atom names, no edits
    if level == 'atomLevel':
        atom_resname = ('%s_%s' % (atom_resname, atom_name))
        nearest_resname_list = [('%s_%s' % (nearest_resname, 
                nearest_atomName))]
    '''

    ###Inclusion of molecule resid
    if level == 'residLevel_resname':
        atom_resname = atom_resname
        if atom_resname not in waterTuple:
            atom_resname = '%s_%s' % (atom_resname, atom_resid)
        nearest_resname_list = [('%s_%s' % (nearest_resname, 
                nearest_resid))]
        #print(atom_resname, nearest_resname_list)

    ###Inclusion of molecule resid
    if level == 'residLevel_atom':
        atom_resname = ('%s_%s' % (atom_resname, atom_name))
        nearest_resname_list = [('%s_%s' % (nearest_resname, 
                nearest_resid))]


    if level == 'soluteContacts':
        atom_resname = atom_resname
        if atom_resname in waterTuple:
            if RAD_nAtoms != None:
                name_id_list = []
                for nInfo in RAD_nAtoms:
                    nResid = nInfo[1]
                    nResname = nInfo[2]

                    if nResname not in waterTuple:
                        name_id = ('%s_%s' % (nResname, nResid))
                        name_id_list.append(name_id)

                name_id_list = sorted(name_id_list)

                if len(name_id_list) > 1:
                    for solute in name_id_list:
                        list2 = name_id_list[:]
                        list2.remove(solute)
                        for solute2 in list2:
                            if ('%s_%s' % (solute, solute2)) not in \
                                    nearest_resname_list:
                                nearest_resname_list.append(
                                        ('%s_%s' % (solute, solute2)))


    return atom_resname, nearest_resname_list, waterTuple





